from pytest import approx, fixture
import random
from solcore import material
from solcore.poisson_drift_diffusion.DeviceStructure import CreateDeviceStructure
from solcore.structure import Junction, Layer, SolcoreMaterialToStr, Structure, ToSolcoreMaterial, TunnelJunction
from solcore.structure import InLineComposition, ToLayer, ToStructure

# Materials
t = 300
qw_mat_material_name = 'InGaAs'
qw_mat_in = 0.2
qw_mat_structure = {'material': qw_mat_material_name, 'element': 'In', 'fraction': qw_mat_in}
b_mat_material_name = 'GaAsP'
b_mat_p = 0.1
b_mat_structure = {'material': b_mat_material_name, 'element': 'P', 'fraction': b_mat_p}
i_gaas_material_name = 'GaAs'
i_gaas_structure = {'material': i_gaas_material_name}

# Layers
available_roles = ['Barrier', 'Base', 'BSF', 'Emitter', 'Intrinsic', 'Window']
wkt_box = 'POLYGON((0 0, 1 0, 1 1, 0 1, 0 0))'
width1 = random.uniform(1e-9, 1e-8)
role1 = random.choice(available_roles)
width2 = random.uniform(1e-9, 1e-8)
role2 = random.choice(available_roles)
width3 = random.uniform(1e-9, 1e-8)
role3 = random.choice(available_roles)

@fixture
def qw_mat():
    return material(qw_mat_material_name)(T=t, In=qw_mat_in)


@fixture
def b_mat():
    return material(b_mat_material_name)(T=t, P=b_mat_p)


@fixture
def i_gaas():
    return material(i_gaas_material_name)(T=t)


@fixture
def layer1(qw_mat):
    return Layer(width1, qw_mat, role1, wkt_box, new_property='new_property')


@fixture
def layer2(b_mat):
    return Layer(width2, b_mat, role2, wkt_box)


@fixture
def layer3(i_gaas):
    return Layer(width3, i_gaas, role3, wkt_box)


@fixture
def device(layer1, layer2, layer3):
    return CreateDeviceStructure(layers=[layer1, layer2, layer3])


def test_layer_and_junction(qw_mat, b_mat, i_gaas, layer1, layer2, layer3):
    assert layer1.width == width1
    assert layer1.role == role1
    assert layer1.material == qw_mat
    assert layer1.geometry == wkt_box
    assert layer1.__dict__['new_property'] == 'new_property'

    assert layer2.width == width2
    assert layer2.role == role2
    assert layer2.material == b_mat
    assert layer2.geometry == wkt_box

    assert layer3.width == width3
    assert layer3.role == role3
    assert layer3.material == i_gaas
    assert layer3.geometry == wkt_box

    sn = random.uniform(1e6, 1e7)
    sp = random.uniform(1e6, 1e7)
    junction1 = Junction([layer1, layer2, layer3], sn=sn, sp=sp, T=t, kind='PDD')

    assert junction1.__len__() == 3
    assert junction1.__dict__ == {'sn': sn, 'sp': sp, 'T': t, 'kind': 'PDD'}
    assert junction1[0] == layer1
    assert junction1[1] == layer2
    assert junction1[2] == layer3

    tunnel1 = TunnelJunction([layer1, layer2, layer3], sn=sn, sp=sp, T=t, kind='PDD')

    assert tunnel1.__len__() == 3
    assert tunnel1.__dict__ == {'sn': sn, 'sp': sp, 'T': t, 'kind': 'PDD', 'R': 1e-16, 'pn': True}
    assert tunnel1[0] == layer1
    assert tunnel1[1] == layer2
    assert tunnel1[2] == layer3


def test_structure(b_mat, layer1, layer2, layer3):
    structure1 = Structure([layer1, layer2], substrate=b_mat, T=t)

    assert structure1.__len__() == 2
    assert structure1.__dict__ == {'substrate': b_mat, 'T': t, 'labels': [None, None]}
    assert structure1.width() == approx(width1 + width2)
    assert structure1[0] == layer1
    assert structure1[1] == layer2

    structure2 = Structure([], substrate=b_mat, T=t)
    structure2.append(layer1, layer_label='layer1')

    assert structure2.__len__() == 1
    assert structure2.__dict__ == {'substrate': b_mat, 'T': t, 'labels': ['layer1']}
    assert structure2[0] == layer1

    structure2.append_multiple([layer2, layer3], layer_labels=['layer2', 'layer3'])

    assert structure2.__len__() == 3
    assert structure2.__dict__ == {'substrate': b_mat, 'T': t, 'labels': ['layer1', 'layer2', 'layer3']}
    assert structure2[0] == layer1
    assert structure2[1] == layer2
    assert structure2[2] == layer3
    assert structure2.width() == approx(width1 + width2 + width3)
    assert structure2.relative_widths()['layer1'] == approx(width1 / structure2.width())
    assert structure2.relative_widths()['layer2'] == approx(width2 / structure2.width())
    assert structure2.relative_widths()['layer3'] == approx(width3 / structure2.width())

    structure3 = Structure([])
    structure3.append(layer1, layer_label='layer1', repeats=2)

    assert structure3.__len__() == 2
    assert structure3.__dict__ == {'labels': ['layer1', 'layer1']}
    assert structure3[0] == layer1
    assert structure3[1] == layer1
    assert structure3.width() == approx(width1 * 2)
    # Below currently fails due to a bug when specifying labels and repeating more than once
    # assert structure3.relative_widths()['layer1'] == width1 / structure3.width()

    structure4 = Structure([])
    structure4.append_multiple([layer1, layer2], layer_labels=['layer1', 'layer2'], repeats=2)

    assert structure4.__len__() == 4
    assert structure4.__dict__ == {'labels': ['layer1', 'layer2', 'layer1', 'layer2']}
    assert structure4[0] == layer1
    assert structure4[1] == layer2
    assert structure4[2] == layer1
    assert structure4[3] == layer2
    assert structure4.width() == approx(width1 * 2 + width2 * 2)
    # Below currently fails due to a bug when specifying labels and repeating more than once
    # assert structure4.relative_widths()['layer1'] == width1 / structure4.width()
    # assert structure4.relative_widths()['layer2'] == width2 / structure4.width()


def test_material_to_str(qw_mat, b_mat, i_gaas):
    assert SolcoreMaterialToStr(qw_mat) == qw_mat_structure
    assert SolcoreMaterialToStr(b_mat) == b_mat_structure
    assert SolcoreMaterialToStr(i_gaas) == i_gaas_structure


def test_to_material(qw_mat, b_mat, i_gaas):
    qw_material = ToSolcoreMaterial(qw_mat_structure, t, True)
    assert qw_material.__class__.__name__ == qw_mat_material_name
    assert qw_material.__dict__ == qw_mat.__dict__

    b_material = ToSolcoreMaterial(b_mat_structure, t, True)
    assert b_material.__class__.__name__ == b_mat_material_name
    assert b_material.__dict__ == b_mat.__dict__

    i_gaas_material = ToSolcoreMaterial(i_gaas_structure, t, True)
    assert i_gaas_material.__class__.__name__ == i_gaas_material_name
    assert i_gaas_material.__dict__ == i_gaas.__dict__


def test_inline_composition(device):
    assert InLineComposition(device['layers'][0]) == 'In0.2GaAs'
    assert InLineComposition(device['layers'][1]) == 'GaAsP0.1'
    assert InLineComposition(device['layers'][2]) == 'GaAs'


def test_to_layer(qw_mat, device):
    device_layer = device['layers'][0]
    composition = device_layer['properties']['composition']
    layer = ToLayer(width1, ToSolcoreMaterial(composition, t), role1)

    assert layer.width == width1
    assert layer.role == role1
    assert layer.material.__str__() == qw_mat.__str__()


def test_to_structure(device):
    structure = ToStructure(device)

    assert structure.__len__() == 3
    assert structure.width() == approx(width1 + width2 + width3)



import numpy as np
import math
from scipy.optimize import root

from solcore.state import State
from solcore.constants import kb, q, c, h


def iv_multijunction(solar_cell, options):
    """ Calculates the overall IV characteristics of any number of junctions numerically at the requested voltage
    points. If photocurrent is not provided, the resultant IV characteristics are purely recombination currents,
    otherwise light IVs are returned.

    In the end, the soalr_cell object is updated with an "iv" attribute containing a dictionary with:
        "IV": (V, I) Calculated IV characteristics
        "junction IV": [(V junc 1, I junc 1), (V junc 2, I junc 2), ...]
        "Rseries IV": (V, I) Calculated IV characteristics of the series resistance
        "coupled IV": For each junction but for the first one, a list with the coupled current coming form the upper junctions.
        "Isc", Voc", "P", "FF" and "Eta": In case of mpp = True and light IV.

    The sign convention is:
        - Photocurrents: Positive.
        - Dark Currents: Negative

    :param solar_cell: A solar cell object with one or more junctions. The IV of the individual junctions must have been calculated already.
    :param kwargs: A dictionary containing, at least, the following elements:
        - mpp: (Boolean) If Isc, Voc, FF, Vmpp, Impp and Pmpp must be calculated.
        - voltages: Array of voltages in which to calculate the data
        - light_iv: (Boolean) if light IV is being calculated
    :return: None

    """
    output_V = options.voltages
    mpp_parameters = options.mpp
    radiative_coupling = options.radiative_coupling
    light_iv = options.light_iv

    Rs = max(solar_cell.R_series, 1e-14)

    # The current and junction voltage arrays
    num_jun = solar_cell.junctions
    tunnel_jun = solar_cell.tunnel_indices
    V_junction_array = np.zeros((len(output_V), num_jun))

    # The following assumes that all junctions have the currents defined at the same voltages
    minimum_J = solar_cell(0).current
    for j in range(1, num_jun):
        minimum_J = np.where(np.absolute(minimum_J) < np.absolute(solar_cell(j).current), minimum_J,
                             solar_cell(j).current)

    minimum_J = np.sort(minimum_J)

    # For each junction, we find the voltage corresponding to each of these currents
    temp_V_junction_array = np.zeros((len(minimum_J), num_jun))
    for j in range(num_jun):
        temp_V_junction_array[:, j] = np.interp(minimum_J, solar_cell(j).current, solar_cell(j).voltage)

    # We calculate the total voltage related to the...
    # ... series resistance
    temp_V_total = Rs * minimum_J

    # ... tunnel junctions
    for j in tunnel_jun:
        temp_V_total -= solar_cell[j].vi(-minimum_J)

    # ... and the normal junctions
    for j in range(num_jun):
        temp_V_total += temp_V_junction_array[:, j]

    # Finally, we calculate the current at the requested voltages and the voltages for each junction
    output_J = np.interp(output_V, temp_V_total, minimum_J)
    for j in range(num_jun):
        V_junction_array[:, j] = np.interp(output_V, temp_V_total, temp_V_junction_array[:, j])

    coupled_J = []
    if radiative_coupling:
        # First of all, we check if all the junctions are DB or 2D
        ok = True
        for i in range(num_jun):
            ok = ok and solar_cell(i).kind in ['DB', '2D']

        if ok:
            output_J, V_junction_array, coupled_J = solve_radiative_coupling(solar_cell, options, V_junction_array)
        else:
            print(
                'WARNING: Only "DB" and "2D" junctions (using the DB solver) can use radiative coupling.\nSkipping calculation...')

    # Finally, we calculate the solar cell parameters
    Isc = None
    Voc = None
    Pmpp = None
    Vmpp = None
    Impp = None
    FF = None
    Eta = None

    # If we are calculating the light IV, we also calculate the main parameters: Jsc, Voc, FF, MPP...
    if mpp_parameters and light_iv:
        VV = output_V
        II = -output_J
        PP = VV * II
        power = options.light_source.power_density
        try:
            Isc = np.interp([0], VV, II)[0]
            Voc = np.interp([0], -II, VV)[0]
            Pmpp = np.max(PP)
            idx = np.argmax(PP)
            Vmpp = VV[idx]
            Impp = II[idx]
            FF = Pmpp / (Isc * Voc)
            Eta = Pmpp/power

        except Exception as err:
            print('Error calculating the MPP parameters: {}'.format(err))

    solar_cell.iv = State({
        "IV": (output_V, -output_J),
        "junction IV": [(V_junction_array[:, i], -output_J) for i in range(num_jun)],
        "coupled IV": coupled_J,
        "Rseries IV": (output_J * Rs, -output_J),
        "Isc": Isc,
        "Voc": Voc,
        "FF": FF,
        "Pmpp": Pmpp,
        "Vmpp": Vmpp,
        "Impp": Impp,
        "Eta": Eta
    })


def solve_radiative_coupling(solar_cell, options, V_junction_array):
    """ Calculates the radiative IV curve of a MJ solar cell in the presence of radiative coupling between subcells.

    WARNING: Tunnel junctions are not implemented in the radiative coupling mode, yet.

    :param solar_cell: The MJ solar cell structure
    :param options: General options of the solver
    :param V_junction_array: Array with all the junction voltages without coupling. Only the first element is used, actually.
    :return: A tupple with J, V_junction_array in the presence of coupling and the coupled current
    """
    print('Calculating IV curve with radiative coupling...')
    print('WARNING: Tunnel junctions are not implemented in the radiative coupling calculator, yet.')

    output_V = options.voltages
    wl = options.wavelength
    T = options.T
    num_jun = solar_cell.junctions

    # First we calculate the front surface reflection and effective refractive index, as we will do normally for a
    # detailed balanced calculation
    Rn = np.minimum(0.999999, solar_cell(0).reflected(wl))

    # Out of the normal incidence reflection we calculate an effective refractive index
    neff = (1 + np.sqrt(Rn)) / (1 - np.sqrt(Rn))

    # And the corresponding critical angle
    cos_tc = np.sqrt(1 - 1 / neff ** 2)
    steps = int(math.ceil(20 / (1 - min(cos_tc))))
    cos_t = np.linspace(0.001, 1, steps)

    # And now, we back calculate the reflection at any internal angle
    R = np.ones((len(wl), steps))

    for i in range(len(wl)):
        idx = np.where(cos_t > cos_tc[i])

        Rs = np.ones(steps)
        Rp = np.ones(steps)

        cos_i = np.sqrt(1 - neff[i] ** 2 * (1 - cos_t[idx] ** 2))

        Rs[idx] = ((cos_i - neff[i] * cos_t[idx]) / (cos_i + neff[i] * cos_t[idx])) ** 2
        Rp[idx] = ((cos_t[idx] - neff[i] * cos_i) / (cos_t[idx] + neff[i] * cos_i)) ** 2

        R[i, :] = (Rs + Rp) / 2

    # Now, we loop over the junctions, considering all possible couplings
    coupling_currents = np.zeros((num_jun - 1, num_jun - 1))

    # The emitting cell
    for i in range(num_jun - 1):
        # B = exp(-alpha*w) but it is not necessary to calculate alpha and w explicitly
        Bem = 1 - solar_cell(i).absorptance(wl)

        n = solar_cell(i).n
        j01 = 2 * q * n ** 2 * c

        # The absorbing cell:
        Btrans = 1
        for j in range(i + 1, num_jun):
            Bab = 1 - solar_cell(j).absorptance(wl)

            A = 0.5 * (1 + R[:, -1] * Bem) * (1 - Bem) * (1 - Bab) * Btrans

            for k in range(1, steps - 1):
                t = cos_t[k]

                # The integral of the absorpton/emission across the back surface
                A = A + (1 + R[:, k] * Bem ** (1 / t)) * (
                    1 - Bem ** (1 / t)) * (1 - Bab ** (1 / t)) * Btrans ** (1 / t) * t

            A = 2 * np.pi * A / (steps + 1)
            boltz = j01 * (A / wl ** 4) * np.exp(- (h * c / wl) / (kb * T))

            # Now the coupling factors are defined, we need to calculate the coupling current, integrating over wavlenegth
            # We use the boltzmann approximation to make our life easier (and faster).
            coupling_currents[i, j - 1] = np.trapz(boltz, wl)

            # And we reduce the fraction of light reaching the next junction
            Btrans = Btrans * Bab

    # With this, we can calculate the IV curves numerically
    R_series = max(solar_cell.R_series, 1e-14)

    # This is the function we want to minimize: the node equations
    def F(Vext, voltages):

        # Create the output
        output = np.zeros_like(voltages)

        # First equation
        I1 = solar_cell(0).iv(voltages[0])
        output[0] = (Vext - np.sum(voltages)) / R_series - I1

        # The rest of the equations
        for j in range(1, num_jun):
            I2 = solar_cell(j).iv(voltages[j])

            # Here is where the coupling comes into play
            for i in range(j):
                I2 = I2 - coupling_currents[i, j - 1] * np.exp((q * voltages[i]) / (kb * T))

            output[j] = I1 - I2

            I1 = I2

        return output

    # Now we loop over all the voltages
    # We get the initial condition from the solution without coupling
    guess = V_junction_array[0]
    output_J = np.zeros_like(output_V)

    for i, VV in enumerate(output_V):
        # We particularise the function to minimize for the current output voltage
        def FF(voltages):
            return F(VV, voltages)

        # Solve the problem
        result = root(FF, guess, method='lm', tol=1e-10)
        V_junction_array[i] = result['x']

        # Finally, we calculate the current, based on the current of the first junction
        output_J[i] = solar_cell(0).iv(result['x'][0])

        guess = V_junction_array[i]

    # We calculate the coupling current and add it to the output
    coupled_J = []
    for j in range(1, num_jun):
        coupled = []
        for i in range(j):
             coupled.append((V_junction_array[:, i], coupling_currents[i, j - 1] * np.exp((q * V_junction_array[:, i]) / (kb * T))))

        coupled_J.append(coupled)

    return output_J, V_junction_array, coupled_J



""" This module contains tools for configuring the solcore installation, adding/removing sources with material data,
etc.
"""
from configparser import ConfigParser
import os
import shutil
from typing import Union, Optional, Callable
from pathlib import Path
from warnings import warn
from collections import defaultdict


SOLCORE_ROOT = str(Path(__file__).parent)


class SolcoreConfig:
    def __init__(self, default_config: Union[Path, str], user_config: Union[Path, str]):
        self.default_config = default_config
        self.user_config = user_config
        self.user_folder = str(Path(user_config).parent)
        self._observers = defaultdict(list)
        self._default_data = ConfigParser()
        self._user_data = ConfigParser()
        self._default_data.optionxform = str
        self._user_data.optionxform = str

        self._default_data.read(self.default_config)
        if not os.path.exists(self.user_config):
            self.reset_defaults(confirm=True)
        else:
            self._user_data.read(self.user_config)

        if (
            "version" not in self.sources("Configuration")
            or self.version() != self._default_data["Configuration"]["version"]
        ):
            self.restore_defaults()

    def reset_defaults(self, confirm: bool = False) -> None:
        """ Resets the default Solcore configuration in the user home folder.

        :return: None
        """
        if not confirm:
            response = input(
                "This action will delete any custom Solcore configuration. "
                "Are you sure you want to continue (Y/n)?"
            )

            if response in "Yy":
                confirm = True

        if confirm:
            shutil.copy2(self.default_config, self.user_config)
            self._user_data.read(self.user_config)

    def _save_user_config(self) -> None:
        """ Saves the current user configuration

        :return: None
        """
        with open(self.user_config, "w") as fp:
            self._user_data.write(fp)

    def register_observer(
        self, section: str, callback: Callable[[str], None]
    ) -> None:
        """ Registers and observer to be called when a section changes.

        :param section: Section to observe.
        :param callback: Function to execute when there are changes. It takes the name
        and value of the updated/new source as inputs.
        :return: None
        """
        if section not in self.sections:
            raise KeyError(f"Unknown section: {section}.")
        self._observers[section].append(callback)

    def _notify_observers(self, section: str, source: str, value: str) -> None:
        """ Notifies the observers of a section of a change in a source.

        :param section: Section changed.
        :param name: Source added/updated.
        :param name: New value
        :return: None
        """
        for c in self._observers[section]:
            c(source=source, value=value)

    def __getitem__(self, item):
        """Gets an item from Solcore's configuration."""
        if isinstance(item, tuple) and len(item) == 2:
            return self._user_data[item[0]][item[1]].replace(
                "SOLCORE_ROOT", SOLCORE_ROOT
            )
        elif isinstance(item, str):
            return self._user_data[item]
        else:
            raise KeyError(f"Invalid config option {item}.")

    def __setitem__(self, key, value):
        """Sets an item in Solcore's configuration."""
        if isinstance(key, tuple) and len(key) == 2 and key[0] in self._user_data:
            self._user_data[key[0]][key[1]] = value
            self._save_user_config()
            self._notify_observers(section=key[0], source=key[1], value=value)
        else:
            raise KeyError(f"Invalid config option {key}.")

    @property
    def sections(self):
        """Sections in the configuration file."""
        return self._user_data.sections()

    def sources(self, section):
        """Sources in each the requested section."""
        return self._user_data.options(section)

    def remove_source(self, section: str, name: str) -> None:
        """ General function to remove sources from the configuration files. It checks if
        the source exists and, if so, if it is a default Solcore source. In the later case,
        it disable the source by setting it to an empty string rather than removing it.

        :param section: The section to remove a source from.
        :param name: The name of the source.
        :return: None
        """
        if name in self._default_data.options(section):
            print(
                f"Source {name} is a default Solcore source. "
                "It can not be removed and will be disabled instead."
            )
            self[section, name] = ""
        else:
            self._user_data.remove_option(section, name)
            self._save_user_config()

        print(f"{section}:{name} source removed!")

    def restore_default_source(self, section: str, name: str) -> None:
        """ Restores the default value of a source, assuming the source has a
        default value.

        :param section: The section of the source.
        :param name: The name of the source.
        :return: None
        """
        if name not in self._default_data.options(section):
            options = self._default_data.options(section)
            print(
                f"There is no default value for {section} source: {name}. Available "
                f"default {section} sources are: {options}"
            )
            return

        self[section, name] = self._default_data[section][name]
        print(f"Default Solcore value for {section} source {name} has been restored!")

    def restore_defaults(self) -> None:
        """Restores all the default values without touching user additions.

        :return: None
        """
        for s in self._default_data.sections():
            for o in self._default_data.options(s):
                self._user_data[s][o] = self._default_data[s][o]
        self._save_user_config()

    def version(self) -> str:
        """ Provides the Solcore version

        :return: The version number
        """
        return self["Configuration", "version"]

    def welcome_message(self, show: Optional[bool] = None) -> bool:
        """ Sets if the welcome message must be shown or not

        :param show: True/False for showing/hiding the welcome message
        :return: None
        """
        if show is not None:
            self["Configuration", "welcome_message"] = str(show)
        return self["Configuration", "welcome_message"] == "True"

    def verbose_loading(self, show: Optional[bool] = None) -> bool:
        """ Sets if the loading messages (besides the welcome message) must be shown
        or not

        :param show: True/False for showing/hiding the loading messages
        :return: None
        """
        if show is not None:
            self["Configuration", "verbose_loading"] = str(show)
        return self["Configuration", "verbose_loading"] == "True"

    def spice(self, location: Optional[Union[Path, str]] = None) -> str:
        """ Sets or returns the location of the SPICE executable.

        :param location: The location of the spice executable.
        :return: The currently configured location for the executable
        """
        if location is not None:
            self["External programs", "spice"] = str(location)
        return self["External programs", "spice"]

    def smarts(self, location: Optional[Union[Path, str]] = None) -> str:
        """ Sets or returns the location of the SMARTS executable.

        :param location: The location of the SMARTS distribution.
        :return: The currently configured location for the executable
        """
        if location is not None:
            self["External programs", "smarts"] = str(location)
        return self["External programs", "smarts"]

    def units(
        self, name: Optional[str] = None, location: Optional[Union[Path, str]] = None
    ) -> Union[list, str]:
        """ Adds a Units source to Solcore or returns the value of an existing one.

        If called without arguments, it returns the list of available Unit sources.

        :param name: The name of the source.
        :param location: The full path to the location of the source. The source must
        be a ConfigParser formatted file.
        :return: The path to the source or a list of available sources
        """
        if name is None:
            return self.sources("Units")
        if location is not None:
            self["Units", name] = str(location)
        return self["Units", name]

    def parameters(
        self, name: Optional[str] = None, location: Optional[Union[Path, str]] = None
    ) -> Union[list, str]:
        """ Adds a Parameters source to Solcore or returns the value of an existing one.

        If called without arguments, it returns the list of available Parameter sources.

        :param name: The name of the source.
        :param location: The full path to the location of the source. The source must
        be a ConfigParser formatted file.
        :return: The path to the source or a list of available sources
        """
        if name is None:
            return self.sources("Parameters")
        if location is not None:
            self["Parameters", name] = str(location)
        return self["Parameters", name]

    def materials(
        self, name: Optional[str] = None, location: Optional[Union[Path, str]] = None
    ) -> Union[list, str]:
        """ Adds a Materials source to Solcore or returns the value of an existing one.

        If called without arguments, it returns the list of available Material sources.

        :param name: The name of the source.
        :param location: The full path to the location of the source. The source must
        be a folder containing the n and k
        data of the material. See the the Material System in the manual for more
        information.
        :return: The path to the source or a list of available sources
        """
        if name is None:
            return self.sources("Materials")
        if location is not None:
            self["Materials", name] = str(location)
        return self["Materials", name]

    def __repr__(self) -> str:
        """ Returns the current Solcore configuration

        :return: String with the configuration
        """
        result = ""
        for section in self._user_data.sections():
            result += f"[{section}]\n"
            for option in self._user_data.options(section):
                result += f"{option} = {self._user_data[section][option]}\n"
            result += "\n"
        return result


def reset_defaults(confirm: bool = False) -> None:
    """ Resets the default Solcore configuration in the user home folder.

    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.reset_defaults instead.",
        FutureWarning,
    )
    from . import config

    config.reset_defaults(confirm=confirm)


def save_user_config() -> None:
    """ Saves the current user configuration

    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.save_user_config instead.",
        FutureWarning,
    )
    from . import config

    config._save_user_config()


def remove_source(section: str, name: str) -> None:
    """ General function to remove sources from the configuration files. It checks if
    the source exists and, if so, if it is a default Solcore source. In the later case,
    it disable the source by setting it to an empty string rather than removing it.

    :param section: The section to remove a source from.
    :param name: The name of the source.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.remove_source instead.",
        FutureWarning,
    )
    from . import config

    config.remove_source(section=section, name=name)


def add_source(section: str, name: str, location: Union[Path, str]) -> None:
    """ General function to add sources to the configuration files.
    If the source already exists, its value will be replaced by the new one.

    :param section: The section to add a source to.
    :param name: The name of the source.
    :param location: Path to the source (file or folder)
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use 'solcore.config[section, option]=value' instead.",
        FutureWarning,
    )
    from . import config

    config.add_source(section=section, name=name, location=location)


def restore_default_source(section: str, name: str) -> None:
    """ Restores the default value of a source, assuming the source has a
    default value.

    :param section: The section of the source.
    :param name: The name of the source.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.restore_default_source instead.",
        FutureWarning,
    )
    from . import config

    config.restore_default_source(section=section, name=name)


def add_units_source(name: str, location: str) -> None:
    """ Adds a Units source to Solcore.

    :param name: The name of the source.
    :param location: The full path to the location of the source.
    The source must be a ConfigParser formatted file.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.units instead.",
        FutureWarning,
    )
    from . import config

    config.units(name, location=location)


def add_parameters_source(name: str, location: str) -> None:
    """ Adds a Parameters source to Solcore.

    :param name: The name of the source.
    :param location: The full path to the location of the source.
    The source must be a ConfigParser formatted file.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.parameters instead.",
        FutureWarning,
    )
    from . import config

    config.parameters(name, location=location)


def add_materials_source(name: str, location: str) -> None:
    """ Adds a Materials source to Solcore.

    :param name: The name of the source.
    :param location: The full path to the location of the source. The source must be a folder containing the n and k
    data of the material. See the the Material System in the manual for more information.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.materials instead.",
        FutureWarning,
    )
    from . import config

    config.materials(name, location=location)


def remove_units_source(name: str) -> None:
    """ Removes a Units source from Solcore.

    :param name: The name of the source.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.remove_source instead.",
        FutureWarning,
    )
    from . import config

    config.remove_source("Units", name)


def remove_parameters_source(name: str) -> None:
    """ Removes a Parameters source from Solcore.

    :param name: The name of the source.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.remove_source instead.",
        FutureWarning,
    )
    from . import config

    config.remove_source("Parameters", name)


def remove_materials_source(name: str) -> None:
    """ Removes a Materials source from Solcore.

    :param name: The name of the source.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.remove_source instead.",
        FutureWarning,
    )
    from . import config

    config.remove_source("Materials", name)


def restore_default_units_source(name: str) -> None:
    """ Restores the default value of a Units source from Solcore.

    :param name: The name of the source.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.restore_default_source instead.",
        FutureWarning,
    )
    from . import config

    config.restore_default_source("Units", name)


def restore_default_parameters_source(name: str) -> None:
    """ Restores the default value of a Parameters source from Solcore.

    :param name: The name of the source.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.restore_default_source instead.",
        FutureWarning,
    )
    from . import config

    config.restore_default_source("Parameters", name)


def restore_default_materials_source(name: str) -> None:
    """ Restores the default value of a Materials source from Solcore.

    :param name: The name of the source.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.restore_default_source instead.",
        FutureWarning,
    )
    from . import config

    config.restore_default_source("Materials", name)


def welcome_message(show: bool) -> None:
    """ Sets if the welcome message must be shown or not

    :param show: True/False for showing/hiding the welcome message
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.welcome_message instead.",
        FutureWarning,
    )
    from . import config

    config.welcome_message(show)


def verbose_loading(show: bool) -> None:
    """ Sets if the loading messages (besides the welcome message) must be shown or not

    :param show: True/False for showing/hiding the loading messages
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.verbose_loading instead.",
        FutureWarning,
    )
    from . import config

    config.verbose_loading(show)


def set_location_of_spice(location: Union[Path, str]) -> None:
    """ Sets the location of the spice executable. It does not test if it works.

    :param location: The location of the spice executable.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.set_location_of_spice instead.",
        FutureWarning,
    )
    from . import config

    config.set_location_of_spice(location)


def set_location_of_smarts(location: Union[Path, str]) -> None:
    """ Sets the location of the SMARTS distribution (the root folder). It does not test if it works.

    :param location: The location of the SMARTS distribution.
    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.set_location_of_smarts instead.",
        FutureWarning,
    )
    from . import config

    config.set_location_of_smarts(location)


def get_current_config() -> None:
    """ Prints the current Solcore configuration

    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use print(solcore.config) instead.",
        FutureWarning,
    )
    from . import config

    print(config)


def check_user_config() -> None:
    """ Checks if there's a user configuration file, asking if it needs to be created.

    :return: None
    """
    warn(
        "Deprecated: this function is deprecated and will be removed in the future."
        "Use solcore.config.check_user_config instead.",
        FutureWarning,
    )
    from . import config

    config.check_user_config()



from scipy.optimize import bisect
from numpy import sqrt, log, pi, linspace, array, ones
from solcore import convert, asUnit
from solcore import ParameterSystem
from solcore.science_tracker import science_reference


# Zero stress criterion from:
# Ekins-Daukes et al. Strain-balanced Criteria for Multiple Quantum Well Structures and Its
# Signature in X-ray Rocking Curves. Crystal growth & design (2002) vol. 2 (4) pp. 287-292

# Critical Thickness from:
# Matthews, J., & Blakeslee, A. (1974). Defects in epitaxial multilayers: I. Misfit dislocations.
# Journal of Crystal Growth, 27, 118-125.


def optimise(well_material, well_thickness, well_fraction, barrier_material, barrier_thickness, barrier_fraction,
        lattice_material, lattice_material_fraction=0, parameter_to_optimise="well_thickness", T=300):
    """ This can't handle sol-core materials yet. Needs layers."""

    options = ["barrier_thickness", "well_thickness", "barrier_fraction", "well_fraction"]
    if parameter_to_optimise not in options:
        print('ERROR: Wring option in strain_balancing.optimise. '
              'The only valid options are: {}, {}, {}, {}'.format(*options))

    science_reference("strain-balancing",
                      "Ekins-Daukes et al. Strain-balanced Criteria for Multiple Quantum Well Structures and Its "
                      "Signature in X-ray Rocking Curves. Crystal growth & design (2002) vol. 2 (4) pp. 287-292")

    get_vurgaftman_parameter = ParameterSystem().get_parameter
    t1 = well_thickness
    t2 = barrier_thickness

    x1 = well_fraction
    x2 = barrier_fraction

    C11_well = get_vurgaftman_parameter(well_material, "c11", x=x1, T=T)
    C12_well = get_vurgaftman_parameter(well_material, "c12", x=x1, T=T)
    A1 = C11_well + C12_well - 2 * C12_well ** 2 / C11_well
    a1 = get_vurgaftman_parameter(well_material, "lattice_constant", x=x1, T=T)

    C11_barrier = get_vurgaftman_parameter(barrier_material, "c11", x=x2, T=T)
    C12_barrier = get_vurgaftman_parameter(barrier_material, "c12", x=x2, T=T)
    A2 = C11_barrier + C12_barrier - 2 * C12_barrier ** 2 / C11_barrier
    a2 = get_vurgaftman_parameter(barrier_material, "lattice_constant", x=x2, T=T)

    a0 = get_vurgaftman_parameter(lattice_material, "lattice_constant", x=lattice_material_fraction, T=T)

    a0 = convert(a0, "angstrom", "m")
    a1 = convert(a1, "angstrom", "m")
    a2 = convert(a2, "angstrom", "m")
    if parameter_to_optimise == "barrier_thickness":
        t2 = - (A1 * t1 * a2 ** 2 * (a0 - a1)) / (A2 * a1 ** 2 * (a0 - a2))
    if parameter_to_optimise == "well_thickness":
        t1 = - (A2 * t2 * a1 ** 2 * (a0 - a2)) / (A1 * a2 ** 2 * (a0 - a1))

    if parameter_to_optimise == "well_fraction":
        def func(x):
            C11_well = get_vurgaftman_parameter(well_material, "c11", x=x, T=T)
            C12_well = get_vurgaftman_parameter(well_material, "c12", x=x, T=T)
            A1 = C11_well + C12_well - 2 * C12_well ** 2 / C11_well
            a1 = get_vurgaftman_parameter(well_material, "lattice_constant", x=x, T=T)
            a1 = convert(a1, "angstrom", "m")

            misfit = - (A1 * t1 * a2 ** 2 * (a0 - a1)) / (A2 * a1 ** 2 * (a0 - a2)) - t2
            return misfit

        print(func(linspace(0, 1, 10)))

        x1 = bisect(func, 0, 1)

    if parameter_to_optimise == "barrier_fraction":
        def func(x):
            C11_barrier = get_vurgaftman_parameter(barrier_material, "c11", x=x, T=T)
            C12_barrier = get_vurgaftman_parameter(barrier_material, "c12", x=x, T=T)
            A2 = C11_barrier + C12_barrier - 2 * C12_barrier ** 2 / C11_barrier
            a2 = get_vurgaftman_parameter(barrier_material, "lattice_constant", x=x, T=T)
            a2 = convert(a2, "angstrom", "m")

            if a0 == a2 and x == 0:
                a2 = get_vurgaftman_parameter(barrier_material, "lattice_constant", x=0.001, T=T)
                a2 = convert(a2, "angstrom", "m")

            misfit = - (A1 * t1 * a2 ** 2 * (a0 - a1)) / (A2 * a1 ** 2 * (a0 - a2)) - t2
            return misfit

        x2 = bisect(func, 0, 1)

    return {
        "well_material": well_material,
        "well_thickness": t1,
        "well_fraction": x1,
        "barrier_material": barrier_material,
        "barrier_thickness": t2,
        "barrier_fraction": x2,
        "lattice_material": lattice_material,
        "lattice_fraction": lattice_material_fraction
    }


def critical_thickness(layer_material="GaInAs", lattice_material="GaAs", layer_fraction=0, lattice_fraction=None, T=300,
                       during_growth=True, final_unit="m", bowed_material="In"):
    science_reference("critical thickness",
                      "Matthews, J., & Blakeslee, A. (1974). Defects in epitaxial multilayers: I. Misfit dislocations. "
                      "Journal of Crystal Growth, 27, 118-125.")
    get_vurgaftman_parameter = ParameterSystem().get_parameter

    other_params = {
        "T": T,
        bowed_material: layer_fraction,
    }
    c11 = get_vurgaftman_parameter(layer_material, "c11", **other_params)
    c12 = get_vurgaftman_parameter(layer_material, "c12", **other_params)
    a = get_vurgaftman_parameter(layer_material, "lattice_constant", **other_params)
    a_lattice = get_vurgaftman_parameter(lattice_material, "lattice_constant", **other_params)

    b = array(a / sqrt(2))
    v = array(abs(c12 / (c11 + c12)))
    f = array(abs(a / a_lattice - 1))

    longest = max([each.size for each in [v, b, f]])

    if longest != 1:
        if b.size != longest:
            assert b.size == 1, "array arguments do not match length"
            b = ones(longest) * b
        if v.size != longest:
            assert v.size == 1, "array arguments do not match length"
            v = ones(longest) * v
        if f.size != longest:
            assert f.size == 1, "array arguments do not match length"
            f = ones(longest) * f
    else:
        v, b, f = [v], [b], [f]

    result = []
    for vi, fi, bi in zip(v, f, b):
        roots_at_soln = lambda hc_over_b: (hc_over_b) / (log(hc_over_b) + 1) - (1 - vi / 4) / (
            fi * pi * (1 + vi))  # matthews & blakeslee
        try:
            hc = bisect(roots_at_soln, 1e-10, 1e10) * bi
        except ValueError:
            raise ValueError("Critical thickness infinite?")

        result.append(convert(hc / 4, "Ang", final_unit)) if during_growth else result.append(
            convert(hc, "Ang", final_unit))
    return array(result)


if __name__ == "__main__":
    T = linspace(300, 1000, 10)
    from solcore.graphing import *

    result = critical_thickness(layer_material="GaInAs", lattice_material="GaAs", layer_fraction=0.24, T=T,
                                final_unit="nm")
    g = Graph(
        GraphData(T, asUnit(result, 'nm'), label="Matthews-Blakeslee $\\frac{h_c}{4}$"),  # , "-", 1,"red"),
        # yscale="log", 
        # xlim=(0.03,0.53), 
        # ylim=(1,1000), 
        xlabel="Temperature (K)",
        ylabel="Critical Thickness (nm)",
        # grid=True,
        title="InGaAs critical thickness"
        # labels = ([(x_at_024,y_at_024,"x=0.24, $\\frac{h_c}{4}$=%.2fnm"%y_at_024)])
    )
    g.draw()
    T = 300
    x = linspace(0, 1, 100)[1:]
    result = critical_thickness(layer_material="GaInAs", lattice_material="GaAs", layer_fraction=x, T=T,
                                final_unit="nm")
    print(result)
    g = Graph(
        GraphData(x, asUnit(result, 'nm'), label="Matthews-Blakeslee $\\frac{h_c}{4}$"),
        # (f[0], f[1], "Previous calculation", "-g")
        # yscale="log",
        # size="4 square",
        # figsize=(4, 8),
        xlim=(0.03, 0.53),
        ylim=(0, 10),
        xlabel="Indium Fraction",
        ylabel="Critical Thickness (nm)",
        # grid=True,
        title="InGaAs critical thickness"
        # labels = ([(x_at_024,y_at_024,"x=0.24, $\\frac{h_c}{4}$=%.2fnm"%y_at_024)])
    )
    g.draw()



"""This module gives access to the vast library of optical constant data, made freely available by the SOPRA-SA
optoelectronics company founded in 1948. For further detail on the data and SOPRA-SA see the legacy website:
http://www.sspectra.com/sopra.html"""

import numpy as np
import os, sys
import re
from natsort import natsorted
from configparser import ConfigParser
from solcore.science_tracker import science_reference
from solcore import config, SOLCORE_ROOT

SOPRA_PATH = os.path.abspath(config['Others']['sopra'].replace('SOLCORE_ROOT', SOLCORE_ROOT))
compounds_path = os.path.join(SOPRA_PATH, "compounds.txt")
compounds_info = ConfigParser()
compounds_info.read(compounds_path)


# Defining the SOPRA_DB class variable
class sopra_database:
    """Import the SOPRA_DB module from the solcore.material_system package and get started by selecting a material from
    the extensive list that SOPRA-SA compiled;

    >>> GaAs = sopra_database('GaAs')

    Once imported a number of useful methods can be called to return n, k and alpha data for the desired material.
    """

    science_reference("All optical constant data made avaialble by SOPRA-SA",
                      "http://www.sspectra.com/sopra.html")

    def __init__(self, Material):
        # Define filepath to the SOPRA database for file import...
        self.__SOPRA_PATH = SOPRA_PATH

        # Load in SOPRA_DB.csv database file
        DB = np.genfromtxt(os.path.join(self.__SOPRA_PATH, "SOPRA_DB_Updated.csv"), delimiter=",", dtype=str)

        self.__fname = None
        for fname, symbol, range, info in DB:

            if re.fullmatch(Material.upper(), fname) is not None:
                self.__fname = fname

                self.path = os.path.join(self.__SOPRA_PATH, self.__fname + ".MAT")

                # self.info contains all detail loaded from SOPRA_DB,csv file...
                self.info = {"Material": symbol,
                             "Wavelength (nm)": range,
                             "File Info": info,
                             "File Path": self.path}

        # If the material name is incorrect then the material attribute is not written to
        if self.__fname is None:

            print("SOPRA_DB :: ERROR :: Material not found in SOPRA_DB... Check materials list...")
            print("Similar Matches ::")
            for fname, symbol, range, info in DB:

                if re.match(Material.upper(), fname) is not None:
                    print(fname)

            # If the exception is caught, exit the program as nothing else useful can be done...
            # sys.exit()
            raise SOPRAError("Material not found in SOPRA database: {}".format(Material))

    @staticmethod
    def material_list():
        """ SOPRA_DB.material_list() :: Loads a list (.pdf file) of all available SOPRA materials. """

        print("Opening List of Available Materials in the SOPRA database")

        # Need different treatment depending on computer OS.
        if sys.platform == 'darwin':
            # Find spaces in the filename and add a \ before (for unix based systems)
            directory = SOPRA_PATH.split(" ")

            new_path = directory[0]
            for i in range(1, len(directory), 1):
                new_path = new_path + "\ " + directory[i]

            os.system("open " + os.path.join(new_path, "List_Of_Files_Updated_PDF.pdf"))

        elif sys.platform == 'linux':
            # Find spaces in the filename and add a \ before (for unix based systems)
            directory = SOPRA_PATH.split(" ")

            new_path = directory[0]
            for i in range(1, len(directory), 1):
                new_path = new_path + "\ " + directory[i]

            os.system("xdg-open " + os.path.join(new_path, "List_Of_Files_Updated_PDF.pdf"))

        elif sys.platform == 'win32':
            # Find spaces in the filename and add a \ before (for unix based systems)

            os.system("start " + os.path.join(SOPRA_PATH, "List_Of_Files_Updated_PDF.pdf"))

    def load_n(self, Lambda=None):
        """ SOPRA_DB.load_n(Lambda) :: Load refractive index (n) data of the requested material.
            Optional argument Lambda allows user to specify a custom wavelength range. data will be interpolated into
                this range before output.

            Returns: Tuple of (Wavelength, n)"""

        try:
            os.stat(self.path)
        except FileNotFoundError:
            print('load_n :: WARNING :: There is no individual data file for, ' + self.material + ".")
            print('This material may be part of a set of varying composition, check the materials list...')
            sys.exit()

        # Load in data from file...
        Wav, n = np.genfromtxt(self.path, delimiter="*", skip_header=3, skip_footer=3, usecols=(2, 3), unpack=True)

        if Lambda is not None:
            # Interpolate in range specified by Lambda...
            n_interp = np.interp(Lambda, Wav, n)

            return (Lambda, n_interp)

        else:
            return (Wav, n)

    def load_k(self, Lambda=None):
        """ SOPRA_DB.load_k(Lambda) :: Load refractive index (n) data of the requested material.
            Optional argument Lambda allows user to specify a custom wavelength range. data will be interpolated into
                this range before output.

            Returns: Tuple of (Wavelength, k) """

        try:
            os.stat(self.path)
        except FileNotFoundError:
            print('load_k :: WARNING :: There is no individual data file for, ' + self.material + ".")
            print('This material may be part of a set of varying composition, check the materials list...')
            sys.exit()

        # Load in data from file...
        Wav, k = np.genfromtxt(self.path, delimiter="*", skip_header=3, skip_footer=3, usecols=(2, 4), unpack=True)

        if Lambda is not None:
            # Interpolate in range specified by Lambda...
            k_interp = np.interp(Lambda, Wav, k)

            return (Lambda, k_interp)
        else:
            return (Wav, k)

    def load_alpha(self, Lambda=None):
        """ SOPRA_DB.load_alpha(Lambda) :: Load refractive index (n) data of the requested material.
            Optional argument Lambda allows user to specify a custom wavelength range. data will be interpolated into
                this range before output.

            Returns: Tuple of (Wavelength, alpha) """

        Wav, k = self.load_k(Lambda=Lambda)

        return (Wav, ((4 * np.pi) / (Wav * 1E-9)) * k)

    def load_temperature(self, Lambda, T=300):
        """ SOPRA_DB.load_temperature(T, Lambda) :: Loads n and k data for a set of materials with temperature dependent
                data sets
            Optional argument T defaults to 300K
            Required argument Lambda specifies a wavelength range and the data is interpolated to fit. This is a
                required argument here as not all data sets in a group are the same length (will be fixed in a
                subsequent update).

            Returns: Tuple of (Wavelength, n, k) """

        T_degC = T - 273.15  # Convert from Kelvin to degC (units given in the data)...

        # Navigate to the correct folder that contains temperature dependent data...
        path = os.path.join(self.__SOPRA_PATH, self.__fname + "_T")

        try:
            os.stat(path)

        except FileNotFoundError:
            print("load_temperature :: WARNING :: Material folder does not exists... Check materials list...")
            # If material folder is not found exit program as nothing more useful can be done...
            sys.exit()

        # if folder exists, read in files from folder...
        Folder = natsorted(os.listdir(path))

        DATA = []
        TEMP = []

        for files in Folder:

            # .DS_Store is a metadata file used in Mac OS X to store various file/ folder info. Ignoring...
            if ".DS_Store" not in files:

                # extract temperature from filename...
                Num = re.findall("[-+]?\d+[\.]?\d*", files)
                Num.append("0")
                TEMP.append(float(Num[0]))

                Wav, n, k = np.genfromtxt(os.path.join(path, files), delimiter="*",
                                          skip_header=3, skip_footer=3, usecols=(2, 3, 4), unpack=True)

                if Lambda is not None:
                    # Interpolate if the Lambda argument is specified, if not pass loaded Wav, n and k...
                    n_interp = np.interp(Lambda, Wav, n)
                    k_interp = np.interp(Lambda, Wav, k)

                    DATA.append((Lambda, n_interp, k_interp, float(Num[0])))

                else:
                    DATA.append((Wav, n, k, float(Num[0])))

        # Check and see if the entered temperature is within the range of data...
        if T_degC <= min(TEMP):
            print("load_temperature :: WARNING :: Desired Temperature < than the minimum (%6.1f K)" % (
                min(TEMP) + 273.15))
            print("Returned interpolated data will be that at Tmin = %6.1f K" % (min(TEMP) + 273.15))

        elif T_degC >= max(TEMP):
            print("load_temperature :: WARNING :: Desired Temperature > than the maximum (%6.1f K)" % (
                max(TEMP) + 273.15))
            print("Returned interpolated data will be that at Tmax = %6.1f K" % (max(TEMP) + 273.15))

        # use linear interpolation to interpolate the data at the desired temperature...
        n_interp_data = []
        k_interp_data = []

        # In range of all wavelenghs...
        for i in range(0, len(DATA[0][0])):

            T_list = []
            k_at_T = []
            n_at_T = []

            # At each wavelenth, build a list of n and k at each T...
            for X, n, k, temp in DATA:
                T_list.append(temp)
                k_at_T.append(k[i])
                n_at_T.append(n[i])

            # Interpolate the point corresponding to T and build the new data array...
            n_interp_data.append(np.interp(T - 273.19, T_list, n_at_T))
            k_interp_data.append(np.interp(T - 273.19, T_list, k_at_T))

        # Return the Wavelength vector and the new n and k data...
        return (DATA[0][0], n_interp_data, k_interp_data)

    def load_composition(self, Lambda, **kwargs):
        """ SOPRA_DB.load_temperature(T, Lambda) :: Loads n and k data for a set of materials with varying composition.
            Required argument Lambda specifies a wavelength range and the data is interpolated to fit. This is a
                required argument here as not all data sets in a group are the same length (will be fixed in a
                subsequent update).
            Keyword argument :: Specify the factional material and fraction of the desired alloy.

            Returns: Tuple of (Wavelength, n, k) """
        # Use of keyword args allows the user to specify the required material fraction for neatness...
        for material in kwargs:
            mat_fraction = material
            frac = kwargs[material]

        # Navigate to the correct folder that contains temperature dependent data...
        path = os.path.join(self.__SOPRA_PATH, self.__fname + "_" + mat_fraction.upper())
        try:
            os.stat(path)

        except FileNotFoundError:
            print("load_composition :: WARNING :: Material folder does not exists... Check materials list or check" +
                  " that composition material is correct...")
            # If material folder is not found exit program as nothing more useful can be done...
            sys.exit()

        # if folder exists, read in files from folder...
        Folder = natsorted(os.listdir(path))

        DATA = []
        COMP = []

        for files in Folder:

            # .DS_Store is a metadata file used in Mac OS X to store various file/ folder info. Ignoring...
            if ".DS_Store" not in files:

                # extract temperature from filename...
                Num = re.findall("[-+]?\d+[\.]?\d*", files)
                Num.append("0")
                COMP.append(float(Num[0]))

                Wav, n, k = np.genfromtxt(os.path.join(path, files), delimiter="*",
                                          skip_header=3, skip_footer=3, usecols=(2, 3, 4), unpack=True)

                if Lambda is not None:
                    # Interpolate if the Lambda argument is specified, if not pass loaded Wav, n and k...
                    n_interp = np.interp(Lambda, Wav, n)
                    k_interp = np.interp(Lambda, Wav, k)

                    DATA.append((Lambda, n_interp, k_interp, float(Num[0])))

                else:
                    DATA.append((Wav, n, k, float(Num[0])))

        # Check and see if the entered temperature is within the range of data...
        if frac <= min(COMP):
            print("load_composition :: WARNING :: Desired composition < than the minimum (%6.1f %%)" % min(COMP))
            print("Returned interpolated data will be that at %6.1f %%" % min(COMP))

        elif frac >= max(COMP):
            print("load_composition :: WARNING :: Desired composition > than the maximum (%6.1f %%)" % max(COMP))
            print("Returned interpolated data will be that at %6.1f %%" % max(COMP))

        # use linear interpolation to interpolate the data at the desired temperature...
        n_interp_data = []
        k_interp_data = []

        # In range of all wavelenghs...
        for i in range(0, len(DATA[0][0])):

            x_list = []
            k_at_C = []
            n_at_C = []

            # At each wavelenth, build a list of n and k at each T...
            for X, n, k, x in DATA:
                x_list.append(x)
                k_at_C.append(k[i])
                n_at_C.append(n[i])

            # Interpolate the point corresponding to T and build the new data array...
            n_interp_data.append(np.interp(frac, x_list, n_at_C))
            k_interp_data.append(np.interp(frac, x_list, k_at_C))

        # Return the Wavelength vector and the new n and k data...
        return (DATA[0][0], n_interp_data, k_interp_data)


class SOPRAError(Exception):
    def __init__(self, message):
        self.message = message



import solcore.science_tracker as st

st.science_reference('GaP and InP refractive index', 'S. Adachi. Optical dispersion relations for GaP, GaAs, GaSb, InP, InAs, InSb, AlxGa1−xAs, and In1−xGaxAsyP1−y, J. Appl. Phys. 66, 6030-6040 (1989)')
st.science_reference('GaInP refractive index', 'M. Schubert, V. Gottschalch, C. M. Herzinger, H. Yao, P. G. Snyder and J. A. Woollam. Optical constants of GaxIn1−xP lattice matched to GaAs, J. Appl. Phys. 77, 3416 (1995)')



""" This module serves as interface between solcore structures (junctions, layers, materials...) and the
transfer matrix package developed by Steven Byrnes and included in the PyPi repository.

"""
import numpy as np
import solcore

from warnings import warn
from collections.abc import Hashable

from solcore.interpolate import interp1d
from solcore.structure import ToStructure
from solcore.absorption_calculator import tmm_core_vec as tmm
from functools import lru_cache, wraps

degree = np.pi / 180


def np_cache(function):
    """This function was taken from https://stackoverflow.com/questions/52331944/cache-decorator-for-numpy-arrays/52332109#52332109
    
    Creates a cacheable version of a function which takes a 1D numpy array as input, by using a wrapping function
    which converts the array to a tuple. It returns a function which returns the same output as the input function,
    but can be cached, avoiding a bottleneck when optical constants in a material are looked up repeatedly.
    
    If the input argument is hashable already - eg. a single float number - then the function is not chached and is returned "as is".

    :param: function: the function of which a cacheable version is to be created
    :return: wrapper: the cacheable version of the function"""
    @lru_cache()
    def cached_wrapper(hashable_array):
        array = np.array(hashable_array)
        return function(array)

    @wraps(function)
    def wrapper(array):
        if isinstance(array, Hashable):
            return function(array)
        else:
            return cached_wrapper(tuple(array))

    # copy lru_cache attributes over too
    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper


class OptiStack(object):
    """ Class that contains an optical structure: a sequence of layers with a thickness and a complex refractive index.

    It serves as an intermediate step between solcore layers and materials and the stack of thicknesses and
    and n and k values necessary to run calculations involving TMM. When creating an OptiStack object, the thicknesses
    of all the layers forming the Solcore structure and the optical data of the materials of the layers are extracted
    and arranged in such a way they can be easily and fastly read by the TMM functions.

    In addition to a solcore structure with Layers, it can also take a list where each element represent a layer
    written as a list and contains the layer thickness and the dielectrical model, the raw n and k data as a function
    of wavelengths, or a whole Device structure as the type used in the PDD model.

    In summary, this class acepts:

        - A solcore structure with layers
        - A list where each element is [thickness, DielectricModel]
        - A list where each element is [thickness, wavelength, n, k]
        - A list mixing the above:
            [ [thickness, DielectricModel],
              [thickness, wavelength, n, k],
              solcore.Layer,
              solcore.Layer ]

    This allows for maximum flexibility when creating the optical model, allowing to construct the stack with
    experimental data, modelled data and known material properties from the database.

    Yet anther way of defining the layers mixes experimental data with a DielectricModel within the same layer but in
    spectrally distinct regions. The syntaxis for the layer is:

    layer = [thickness, wavelength, n, k, DielectricModel, mixing]

    where mixing is a list containing three elements: [the mixing point (nm), the mixing width (nm),  zero or one]
    depending if the mixing function should be increasing with the wavelength or decreasing. If increasing (zero), the
    Dielectric model will be used at long wavelengths and the experimental data at short wavelengths. If decreasing
    (one) the oposite is done. The mixing point and mixing width control how smooth is the transition between one and
    the other type of data.

    Extra layers such as he semi-infinite, air-like first and last medium, and a back highly absorbing layer are
    included at runtime to fulfill the requirements of the TMM solver or to solve some of its limitations.
    """

    def __init__(self, structure=(), no_back_reflection=False, substrate=None, incidence=None, **kwargs):
        """ Class constructor. It takes a Solcore structure and extract the thickness and optical data from the
        Layers and the materials. Option is given to indicate if the reflexion from the back of the structure must be
        supressed, usefull for ellipsometry calculations. This is done by creating an artificial highly absorbing but
        not reflecting layer just at the back.

        Alternativelly, it can take simply a list of [thickness, DielectricModel] or [thickness, wavelength, n, k] for
        each layer accounting for the refractive index of the layers. The three options can be mixed for maximum
        flexibility.

        :param structure: A list with one or more layers.
        :param no_back_reflection: If reflection from the back must be suppressed. Default=False.
        :param substrate: a semi-infinite transmission medium. Note that if no_back_reflection is set to True,
        adding a substrate won't make any difference.
        :param incidence: a semi-infinite transmission medium. Note that add things may happen if
        this material is absorbing
        """

        self.widths = []
        self.n_data = []
        self.k_data = []
        self.models = []
        self.layers = []
        self.substrate = substrate
        self.incidence = incidence

        self.num_layers = 0
        self.add_layers(structure)

        self.no_back_reflection = no_back_reflection

        if 'no_back_reflexion' in kwargs:
            warn('The no_back_reflexion warning is deprecated. Use no_back_reflection instead.', FutureWarning)
            self.no_back_reflection = kwargs['no_back_reflexion']

    def get_indices(self, wl):
        """ Returns the complex refractive index of the stack.

        :param wl: Wavelength of the light in nm.
        :return: A list with the complex refractive index of each layer, including the semi-infinite front and back
        layers and, opionally, the back absorbing layer used to suppress back surface relfexion.
        """

        out = []
        wl_m = solcore.si(wl, 'nm')

        if hasattr(self, 'n_sub'):
            n1 = self.n_sub(wl_m) + self.k_sub(wl_m)*1.0j
        else:
            if hasattr(wl, 'size'):
                n1 = np.ones_like(wl, dtype=complex)
            else:
                n1 = 1

        # incidence medium!
        if hasattr(self, 'n_sup'):
            n0 = self.n_sup(wl_m) #+ self.k_sup(wl_m)*1.0j ignore complex part for now to avoid errors
        else:
            if hasattr(wl, 'size'):
                n0 = np.ones_like(wl, dtype=complex)
            else:
                n0 = 1

        for i in range(self.num_layers):
            out.append(self.n_data[i](wl_m) + self.k_data[i](wl_m) * 1.0j)

        # substrate irrelevant if no_back_reflection = True
        if self.no_back_reflection:
            return [n0] + out + [self.n_data[-1](wl_m) + self._k_absorbing(wl_m) * 1.0j, n1] # look at last entry in stack,
            # make high;y absorbing layer based on it.

        else:

            return [n0] + out + [n1]


    def get_widths(self):
        """ Returns the widths of the layers of the stack.

        :return: A list with the widths each layer, including the semi-infinite front and back layers and, opionally,
        the back absorbing layer used to suppress back surface relfexion, defined as 1 mm thick.
        """

        if self.no_back_reflection:
            return [np.inf] + self.widths + [1e6, np.inf]
        else:
            return [np.inf] + self.widths + [np.inf]

    def _k_absorbing(self, wl):
        """ k value of the back highly absorbing layer. It is the maximum between the bottom layer of the stack or a
        finite, small value that will absorb all light within the absorbing layer thickness.

        :param wl: Wavelength of the light in nm.
        :return: The k value at each wavelength.
        """
        return np.maximum(wl / 1e-3, self.k_data[-1](wl))

    @staticmethod
    def _k_dummy(wl):
        """ Dummy k value to be used with the dielectric model, which produces the refractive index as a complex
        number.

        :param wl: Wavelength of the light in nm.
        :return: The k value at each wavelength... set to zero.
        """
        return 0.

    def add_layers(self, layers):
        """ Generic function to add layers to the OptiStack. Internally, it calls add_solcore_layer,
        add_modelled_layer or add_raw_nk_layer.

        :param layers: A list with the layers to add (even if it is just one layer) It can be one or more and it can
        mixed, Solcore-based and modelled layers.
        :return: None
        """
        try:
            # If the input is a whole device structure, we get just the layers information
            if type(layers) is dict:
                layers = ToStructure(layers)

            if self.substrate is not None:
                self.n_sub = self.substrate.n
                self.k_sub = self.substrate.k

            if self.incidence is not None:
                self.n_sup = self.incidence.n
                self.k_sup = self.incidence.k

            for layer in layers:
                self.layers.append(layer)
                if 'Layer' in str(type(layer)):
                    self._add_solcore_layer(layer)
                elif 'DielectricConstantModel' in str(type(layer[1])):
                    self._add_modelled_layer(layer)
                else:
                    self._add_raw_nk_layer(layer)

                self.num_layers += 1

        except Exception as err:
            print('Error when adding a new layer to the OptiStack. {}'.format(err))

    def remove_layer(self, idx):
        """ Removes layer with index idx from the OptiStack

        :param idx: Index of the layer to remove
        :return: None
        """
        if idx >= self.num_layers:
            print('Error when removing layers. idx must be: 0 <= idx <= {}.'.format(self.num_layers - 1))
            return

        self.widths.pop(idx)
        self.models.pop(idx)
        self.n_data.pop(idx)
        self.k_data.pop(idx)
        self.num_layers -= 1

    def swap_layers(self, idx1, idx2):
        """ Swaps two layers in the OptiStack.

        :param idx1: The index of one of the layers.
        :param idx2: The index of the other.
        :return: None
        """
        if idx1 >= self.num_layers or idx2 >= self.num_layers:
            print('Error when removing layers. idx must be: 0 <= idx1, idx2 <= {}.'.format(self.num_layers - 1))
            return

        self.widths[idx1], self.widths[idx2] = self.widths[idx2], self.widths[idx1]
        self.models[idx1], self.models[idx2] = self.models[idx2], self.models[idx1]
        self.n_data[idx1], self.n_data[idx2] = self.n_data[idx2], self.n_data[idx1]
        self.k_data[idx1], self.k_data[idx2] = self.k_data[idx2], self.k_data[idx1]


    def set_widths(self, widths):
        """Changes the widths of the layers in the stack.

        :param: widths: a list or array of widths, length equal to the number of layers
        :return: None"""

        if type(widths) is np.ndarray:
            widths = widths.tolist()

        assert len(widths) == self.num_layers, \
            'Error: The list of widths must have as many elements (now {}) as the ' \
        'number of layers (now {}).'.format(len(widths), self.num_layers)

        self.widths = widths


    def _add_solcore_layer(self, layer):
        """ Adds a Solcore layer to the end (bottom) of the stack, extracting its thickness and n and k data.

        :param layer: The Solcore layer
        :return: None
        """
        self.widths.append(solcore.asUnit(layer.width, 'nm'))
        self.models.append([])
        self.n_data.append(np_cache(layer.material.n))
        self.k_data.append(np_cache(layer.material.k))


    def _add_modelled_layer(self, layer):
        """ Adds a layer to the end (bottom) of the stack. The layer must be defined as a list containing the layer
        thickness in nm and a dielectric model.

        :param layer: The new layer to add as [thickness, DielectricModel]
        :return: None
        """
        self.widths.append(layer[0])
        self.models.append(layer[1])
        self.n_data.append(np_cache(self.models[-1].n_and_k))
        self.k_data.append(np_cache(self._k_dummy))

    def _add_raw_nk_layer(self, layer):
        """ Adds a layer to the end (bottom) of the stack. The layer must be defined as a list containing the layer
        thickness in nm, the wavelength, the n and the k data as array-like objects.

        :param layer: The new layer to add as [thickness, wavelength, n, k]
        :return: None
        """
        # We make sure that the wavelengths are increasing, revering the arrays otherwise.
        if layer[1][0] > layer[1][-1]:
            layer[1] = layer[1][::-1]
            layer[2] = layer[2][::-1]
            layer[3] = layer[3][::-1]

        self.widths.append(layer[0])

        if len(layer) >= 5:
            self.models.append(layer[4])
            c = solcore.si(layer[5][0], 'nm')
            w = solcore.si(layer[5][1], 'nm')
            d = layer[5][2]  # = 0 for increasing, =1 for decreasing

            def mix(x):

                out = 1 + np.exp(-(x - c) / w)
                out = d + (-1) ** d * 1 / out

                return out

            n_data = np_cache(lambda x: self.models[-1].n_and_k(x) * mix(x) + (1 - mix(x)) * interp1d(
                x=solcore.si(layer[1], 'nm'), y=layer[2], fill_value=layer[2][-1])(x))
            k_data = np_cache(lambda x: interp1d(x=solcore.si(layer[1], 'nm'), y=layer[3], fill_value=layer[3][-1])(x))

            self.n_data.append(n_data)
            self.k_data.append(k_data)

        else:
            self.models.append([])
            self.n_data.append(np_cache(interp1d(x=solcore.si(layer[1], 'nm'), y=layer[2], fill_value=layer[2][-1])))
            self.k_data.append(np_cache(interp1d(x=solcore.si(layer[1], 'nm'), y=layer[3], fill_value=layer[3][-1])))


def calculate_rat(structure, wavelength, angle=0, pol='u',
                  coherent=True, coherency_list=None, no_back_reflection=True, **kwargs):
    """ Calculates the reflected, absorbed and transmitted intensity of the structure for the wavelengths and angles
    defined.

    :param structure: A solcore Structure object with layers and materials or a OptiStack object.
    :param wavelength: Wavelengths (in nm) in which calculate the data. An array.
    :param angle: Angle (in degrees) of the incident light. Default: 0 (normal incidence).
    :param pol: Polarisation of the light: 's', 'p' or 'u'. Default: 'u' (unpolarised).
    :param coherent: If the light is coherent or not. If not, a coherency list must be added.
    :param coherency_list: A list indicating in which layers light should be treated as coeherent ('c') and in which
    incoherent ('i'). It needs as many elements as layers in the structure.
    :param no_back_reflection: If reflexion from the back must be supressed. Default=True.
    :return: A dictionary with the R, A and T at the specified wavelengths and angle.
    """
    num_wl = len(wavelength)

    if 'no_back_reflexion' in kwargs:
        warn('The no_back_reflexion argument is deprecated. Use no_back_reflection instead.', FutureWarning)
        no_back_reflection = kwargs['no_back_reflexion']

    if 'OptiStack' in str(type(structure)):
        stack = structure
        stack.no_back_reflection = no_back_reflection
    else:
        stack = OptiStack(structure, no_back_reflection=no_back_reflection)

    if not coherent:
        if coherency_list is not None:
            assert len(coherency_list) == stack.num_layers, \
                'Error: The coherency list must have as many elements (now {}) as the ' \
                'number of layers (now {}).'.format(len(coherency_list), stack.num_layers)

            if stack.no_back_reflection:
                coherency_list = ['i'] + coherency_list + ['i', 'i']
            else:
                coherency_list = ['i'] + coherency_list + ['i']

        else:
            raise Exception('Error: For incoherent or partly incoherent calculations you must supply the '
                            'coherency_list parameter with as many elements as the number of layers in the '
                            'structure')

    output = {'R': np.zeros(num_wl), 'A': np.zeros(num_wl), 'T': np.zeros(num_wl), 'all_p': [], 'all_s': [], 'out': [],
              'out_s': [], 'out_p': [],  'A_per_layer_s': [], 'A_per_layer_p': []}

    if pol in 'sp':
        if coherent:
            out = tmm.coh_tmm(pol, stack.get_indices(wavelength), stack.get_widths(), angle * degree, wavelength)
            A_per_layer = tmm.absorp_in_each_layer(out)
            output['R'] = out['R']
            output['A'] = 1 - out['R'] - out['T']
            output['T'] = out['T']
            output['A_per_layer'] = A_per_layer
            output['out'] = out
        else:
            out = tmm.inc_tmm(pol, stack.get_indices(wavelength), stack.get_widths(), coherency_list, angle * degree, wavelength)
            A_per_layer = np.array(tmm.inc_absorp_in_each_layer(out))
            output['R'] = out['R']
            output['A'] = 1 - out['R'] - out['T']
            output['T'] = out['T']
            output['A_per_layer'] = A_per_layer
            output['out'] = out
    else:
        if coherent:
            out_p = tmm.coh_tmm('p', stack.get_indices(wavelength), stack.get_widths(), angle * degree, wavelength)
            out_s = tmm.coh_tmm('s', stack.get_indices(wavelength), stack.get_widths(), angle * degree, wavelength)
            A_per_layer_p = tmm.absorp_in_each_layer(out_p)
            A_per_layer_s = tmm.absorp_in_each_layer(out_s)
            output['R'] = 0.5 * (out_p['R'] + out_s['R'])
            output['T'] = 0.5 * (out_p['T'] + out_s['T'])
            output['A'] = 1 - output['R'] - output['T']
            output['A_per_layer'] = 0.5*(A_per_layer_p + A_per_layer_s)

            output['out_s'] = out_s
            output['out_p'] = out_p
            output['A_per_layer_s'] = A_per_layer_s
            output['A_per_layer_p'] = A_per_layer_p

        else:


            out_p = tmm.inc_tmm('p', stack.get_indices(wavelength), stack.get_widths(), coherency_list, angle * degree, wavelength)
            out_s = tmm.inc_tmm('s', stack.get_indices(wavelength), stack.get_widths(), coherency_list, angle * degree, wavelength)

            A_per_layer_p = np.array(tmm.inc_absorp_in_each_layer(out_p))
            A_per_layer_s = np.array(tmm.inc_absorp_in_each_layer(out_s))

            output['R'] = 0.5 * (out_p['R'] + out_s['R'])
            output['T'] = 0.5 * (out_p['T'] + out_s['T'])
            output['A'] = 1 - output['R'] - output['T']
            output['all_p'] = out_p['power_entering_list']
            output['all_s'] = out_s['power_entering_list']
            output['A_per_layer'] = 0.5*(A_per_layer_p + A_per_layer_s)

            output['out_s'] = out_s
            output['out_p'] = out_p
            output['A_per_layer_s'] =  A_per_layer_s
            output['A_per_layer_p'] = A_per_layer_p


    return output


def calculate_ellipsometry(structure, wavelength, angle, no_back_reflection=True, **kwargs):
    """ Calculates the ellipsometric parameters psi and delta. It can only deal with coherent light and the whole stack
    (including back surface) is considered, so caution must be taken when comparing the simulated results with
    experiments where the back surface is rough or layers are thick and coherent light propagation makes no sense.

    The optional argument no_back_reflection can be included to add an extra layer on the back absorbing all light that
    reaches that position without any reflexion, to remove the reflexion from the back surface.

    :param structure: A solcore structure with layers and materials.
    :param wavelength: Wavelengths (in nm) in which calculate the data. An array.
    :param angle: A tupple or list with the angles (in degrees) in which to calculate the data.
    :param no_back_reflection: If reflexion from the back must be suppressed. Default=True.
    :return: A dictionary with psi and delta at the specified wavelengths and angles (2D arrays).
    """

    if 'no_back_reflexion' in kwargs:
        warn('The no_back_reflexion argument is deprecated. Use no_back_reflection instead.', FutureWarning)
        no_back_reflection = kwargs['no_back_reflexion']

    num_wl = len(wavelength)
    num_ang = len(angle)

    if 'OptiStack' in str(type(structure)):
        stack = structure
        stack.no_back_reflection = no_back_reflection
    else:
        if hasattr(structure, 'substrate'):
            substrate = structure.substrate
        else:
            substrate = None
        stack = OptiStack(structure, no_back_reflection=no_back_reflection, substrate=substrate)

    output = {'psi': np.zeros((num_wl, num_ang)), 'Delta': np.zeros((num_wl, num_ang))}

    for i, ang in enumerate(angle):
        out = tmm.ellips(stack.get_indices(wavelength), stack.get_widths(), ang * degree, wavelength)
        output['psi'][:, i] = out['psi'] / degree

        # We revere the sign of Delta in order to use Woollam sign convention
        output['Delta'][:, i] = -out['Delta'] / degree

        output['Delta'][:, i] = np.where(output['Delta'][:, i] > 0, -output['Delta'][:, i], output['Delta'][:, i])
        output['Delta'][:, i] = np.where(output['Delta'][:, i] < 180, 180 - output['Delta'][:, i],
                                         output['Delta'][:, i])

    return output



def calculate_absorption_profile(structure, wavelength, z_limit=None, steps_size=2, dist=None,
                                   no_back_reflection=True, angle=0, pol = 'u',
                                 coherent=True, coherency_list=None, zero_threshold=1e-6, RAT_out=None, **kwargs):
    """ It calculates the absorbed energy density within the material. From the documentation:

    'In principle this has units of [power]/[volume], but we can express it as a multiple of incoming light power
    density on the material, which has units [power]/[area], so that absorbed energy density has units of 1/[length].'

    Integrating this absorption profile in the whole stack gives the same result that the absorption obtained with
    calculate_rat as long as the spacial mesh (controlled by steps_thinest_layer) is fine enough. If the structure is
    very thick and the mesh not thin enough, the calculation might diverge at short wavelengths.

    :param structure: A solcore structure with layers and materials.
    :param wavelength: Wavelengths in which calculate the data (in nm). An array
    :param z_limit: Maximum value in the z direction
    :param steps_size: if the dist is not specified, the step size in nm to use in the depth-dependent calculation
    :param dist: the positions (in nm) at which to calculate depth-dependent absorption
    :param no_back_reflection: whether to suppress reflections from the back interface (True) or not (False)
    :param angle: incidence of angle in degrees
    :param pol: polarization of incident light: 's', 'p' or 'u' (unpolarized)
    :param coherent: True if all the layers are to be treated coherently, False otherwise
    :param coherency_list: if coherent is False, a list of 'c' (coherent) or 'i' (incoherent) for each layer
    :param zero_threshold: when the fraction of incident light absorbed in a layer is less than this value, the absorption
    profile is completely set to zero for both coherent and incoherent calculations. This is applied on a wavelength-by-wavelength
    basis and is intended to prevent errors where integrating a weak absorption profile in a layer over many points leads to
    calculated EQE > total absorption in that layer.
    :param RAT_out: output from calculate_rat for the same stack & options
    :return: A dictionary containing the positions (in nm) and a 2D array with the absorption in the structure as a
    function of the position and the wavelength.
    """

    if 'no_back_reflexion' in kwargs:
        warn('The no_back_reflexion warning is deprecated. Use no_back_reflection instead.', FutureWarning)
        no_back_reflection = kwargs['no_back_reflexion']

    if RAT_out is None:
        # R, A per layer, T not yet calculated, calculate now
        RAT_out = calculate_rat(structure, wavelength, angle, pol=pol, coherent=coherent, coherency_list=coherency_list,
                                no_back_reflection=no_back_reflection)

    num_wl = len(wavelength)

    if 'OptiStack' in str(type(structure)):
        stack = structure
        stack.no_back_reflection = no_back_reflection
    else:
        stack = OptiStack(structure, no_back_reflection=no_back_reflection)

    if dist is None:
        if z_limit is None:
            z_limit = np.sum(np.array(stack.widths))
        dist = np.arange(0, z_limit, steps_size)

    if not coherent:
        if coherency_list is not None:

            if stack.no_back_reflection:
                coherency_list = ['i'] + coherency_list + ['i', 'i']
            else:
                coherency_list = ['i'] + coherency_list + ['i']

        else:
            raise Exception('Error: For incoherent or partly incoherent calculations you must supply the '
                            'coherency_list parameter with as many elements as the number of layers in the '
                            'structure')

    output = {'position': dist, 'absorption': np.zeros((num_wl, len(dist)))}

    if pol in 'sp':

        if coherent:
            A_per_layer = RAT_out['A_per_layer']
            no_abs_in_layer = np.where(A_per_layer[:-1,:] < zero_threshold)
            RAT_out['out']['vw_list'][no_abs_in_layer[0], no_abs_in_layer[1], :] = 0

            layer, d_in_layer = tmm.find_in_structure_with_inf(stack.get_widths(), dist)
            data = tmm.position_resolved(layer, d_in_layer, RAT_out['out'])
            output['absorption'] = data['absor']

        else:
            layer, d_in_layer = tmm.find_in_structure_with_inf(stack.get_widths(), dist)
            data = tmm.inc_position_resolved(layer, d_in_layer, RAT_out['out'], coherency_list,
                                             4*np.pi*np.imag(stack.get_indices(wavelength))/wavelength, zero_threshold)
            output['absorption'] = data

    else:



        if coherent:

            A_per_layer_s = RAT_out['A_per_layer_s']
            A_per_layer_p = RAT_out['A_per_layer_p']
            no_abs_in_layer_s = np.where(A_per_layer_s[:-1,:] < zero_threshold)
            no_abs_in_layer_p = np.where(A_per_layer_p[:-1, :] < zero_threshold)
            RAT_out['out_s']['vw_list'][no_abs_in_layer_s[0], no_abs_in_layer_s[1], :] = 0
            RAT_out['out_p']['vw_list'][no_abs_in_layer_p[0], no_abs_in_layer_p[1], :] = 0

            layer, d_in_layer = tmm.find_in_structure_with_inf(stack.get_widths(), dist)

            data_s = tmm.position_resolved(layer, d_in_layer, RAT_out['out_s'])
            data_p = tmm.position_resolved(layer, d_in_layer, RAT_out['out_p'])

            output['absorption'] = 0.5*(data_s['absor'] + data_p['absor'])

        else:

            layer, d_in_layer = tmm.find_in_structure_with_inf(stack.get_widths(), dist)
            data_s = tmm.inc_position_resolved(layer, d_in_layer, RAT_out['out_s'], coherency_list,
                                             4*np.pi*np.imag(stack.get_indices(wavelength))/wavelength, zero_threshold)
            data_p = tmm.inc_position_resolved(layer, d_in_layer, RAT_out['out_p'], coherency_list,
                                             4*np.pi*np.imag(stack.get_indices(wavelength))/wavelength, zero_threshold)

            output['absorption'] = 0.5*(data_s + data_p)

    return output


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from solcore import material, si
    from solcore.structure import Layer, Structure

    GaAs = material('GaAs')(T=300)
    InGaAs = material('InGaAs')(T=300, In=0.1)

    my_structure = Structure([
        Layer(si(3000, 'nm'), material=InGaAs),
        Layer(si(30, 'um'), material=GaAs),

    ])

    wavelength = np.linspace(450, 1100, 300)

    out = calculate_rat(my_structure, wavelength, coherent=True, no_back_reflection=False)
    # #
    # plt.plot(wavelength, out['R'], 'b', label='Reflexion')
    # plt.plot(wavelength, out['A'], 'r', label='Absorption')
    # plt.plot(wavelength, out['T'], 'g', label='Transmission')
    # plt.legend()
    # plt.show()

    angles = [60, 65, 70]
    out = calculate_ellipsometry(my_structure, wavelength, angle=angles)
    #
    plt.plot(wavelength, out['psi'][:, 0], 'b', label='psi')
    plt.plot(wavelength, out['Delta'][:, 0], 'r', label='Delta')
    for i in range(1, len(angles)):
        plt.plot(wavelength, out['psi'][:, i], 'b')
        plt.plot(wavelength, out['Delta'][:, i], 'r')
    #
    # plt.legend()

    # out = calculate_absorption_profile_2(my_structure, wavelength, z_limit=3000)
    # # print(tuple(out['absorption'][0]))
    # # plt.plot(out['position'], out['absorption'][0])
    # A = np.zeros_like(wavelength)
    # #
    # for i, absorption in enumerate(out['absorption'][:]):
    #     A[i] = np.trapz(absorption, out['position'])
    # #
    # plt.plot(wavelength, A, 'k', label='Integrated Abs')

    #
    # plt.contourf(out['position'], wavelength, out['absorption'], 200)
    plt.legend()
    plt.show()



from pytest import approx
import numpy as np


def test_prepare_solar_cell_basics(prepare_test_cell):

    from solcore.solar_cell_solver import prepare_solar_cell
    from solcore.state import State

    solar_cell, widths = prepare_test_cell

    options = State()
    options.position = 1e-8

    prepare_solar_cell(solar_cell, options=options)

    # check total width
    assert solar_cell.width * 1e9 == approx(np.sum(widths))

    # check individual width calculations

    indiv_width_int = np.array([layer_obj.width for layer_obj in solar_cell]) * 1e9

    indiv_width = np.array(
        [
            widths[0],
            widths[1],
            np.sum(widths[2:5]),
            np.sum(widths[5:7]),
            np.sum(widths[7:9]),
        ]
    )

    assert indiv_width_int == approx(indiv_width)

    # check offset calculations

    offsets_int = np.array([layer_obj.offset for layer_obj in solar_cell]) * 1e9
    offsets = np.insert(
        np.cumsum(
            np.array(
                [
                    widths[0],
                    widths[1],
                    np.sum(widths[2:5]),
                    np.sum(widths[5:7]),
                    np.sum(widths[7:9]),
                ]
            )
        )[0:-1],
        0,
        0,
    )

    assert offsets_int == approx(offsets)

    # check junction kind
    assert solar_cell[2].kind == "DA"
    assert solar_cell[4].kind == "PDD"


def test_prepare_solar_cell_position(prepare_test_cell):
    from solcore.solar_cell_solver import process_position
    from solcore.state import State
    from solcore.structure import Layer, TunnelJunction, Junction

    solar_cell, widths = prepare_test_cell

    options = State()

    offset = 0
    layer_widths = []
    for j, layer_object in enumerate(solar_cell):

        # Independent layers, for example in a AR coating
        if type(layer_object) is Layer:
            layer_widths.append(layer_object.width)

        # Each Tunnel junctions can also have some layers with a given thickness.
        elif type(layer_object) is TunnelJunction:
            junction_width = 0
            for i, layer in enumerate(layer_object):
                junction_width += layer.width
                layer_widths.append(layer.width)
            solar_cell[j].width = junction_width

        # For each junction, and layer within the junction, we get the layer width.
        elif type(layer_object) is Junction:
            junction_width = 0
            for i, layer in enumerate(layer_object):
                layer_widths.append(layer.width)
                junction_width += layer.width
            solar_cell[j].width = junction_width

        solar_cell[j].offset = offset
        offset += solar_cell[j].width

    solar_cell.width = offset

    pos = 1e-8
    options.position = pos
    process_position(solar_cell, options, layer_widths)
    assert options.position == approx(np.arange(0, solar_cell.width, pos))

    pos = [1e-8]
    options.position = pos
    process_position(solar_cell, options, layer_widths)
    assert options.position == approx(np.arange(0, solar_cell.width, pos[0]))

    pos = np.random.rand(len(solar_cell)) * 1e-9
    options.position = pos
    process_position(solar_cell, options, layer_widths)
    assert options.position == approx(
        np.hstack(
            [
                np.arange(
                    layer_object.offset,
                    layer_object.offset + layer_object.width,
                    pos[j],
                )
                for j, layer_object in enumerate(solar_cell)
            ]
        )
    )

    pos = np.random.rand(len(widths)) * 1e-9
    options.position = pos
    process_position(solar_cell, options, layer_widths)
    layer_offsets = np.insert(np.cumsum(widths), 0, 0) * 1e-9
    assert options.position == approx(
        np.hstack(
            [
                np.arange(layer_offsets[j], layer_offsets[j] + layer_width, pos[j])
                for j, layer_width in enumerate(widths * 1e-9)
            ]
        )
    )

    options.position = None
    process_position(solar_cell, options, layer_widths)
    layer_offsets = np.insert(np.cumsum(layer_widths), 0, 0)
    pos = [max(1e-10, width / 5000) for width in layer_widths]
    expected = np.hstack(
        [
            np.arange(layer_offsets[j], layer_offsets[j] + layer_width, pos[j])
            for j, layer_width in enumerate(layer_widths)
        ]
    )
    assert options.position == approx(expected)

    pos = np.arange(0, solar_cell.width, 1e-8)
    options.position = pos
    process_position(solar_cell, options, layer_widths)
    assert np.all(options.position == pos)



import numpy as np
from scipy.integrate import quad


class Epsi1:
    def __init__(self, epsi_2, mini=0., maxi=5.):

        self.epsi_2 = epsi_2
        self.mini = max(mini, 0)
        self.maxi = maxi

    def __call__(self, E):
        eps = 1e-2
        fun = lambda w, q: w * self.epsi_2(w) / (w ** 2 - q ** 2)

        try:
            val = np.zeros_like(E)
            for i, ee in enumerate(E):
                out1 = quad(fun, a=self.mini, b=ee - eps, args=(ee,))
                out2 = quad(fun, a=ee + eps, b=self.maxi, args=(ee,))

                val[i] = out1[0] + out2[0]
        except:
            out1 = quad(fun, a=self.mini, b=E - eps, args=(E,))
            out2 = quad(fun, a=E + eps, b=self.maxi, args=(E,))

            val = out1[0] + out2[0]

        val *= 2 / np.pi

        return val


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from solcore.absorption_calculator.dielectric_constant_models import Lorentz


    def epsi2(x):
        En = 2.5
        Br = 0.5
        A = 1

        s = 0.5 * Br / np.sqrt(np.log(2))

        out1 = np.exp(-((x - En) / s) ** 2)
        out2 = np.exp(-((x + En) / s) ** 2)

        out = A * (out1 - out2)

        return out


    epsi1 = Epsi1(epsi2)

    E = np.linspace(0, 5, 50)

    e1 = epsi1(E)
    e2 = epsi2(E)

    n = Lorentz(0.2, 2.5, 0.5).dielectric(1240 / E)
    e1_l = np.real(n)
    e2_l = np.imag(n)

    # print(e1, e2)

    plt.plot(E, e1, 'b', label='e1')
    plt.plot(E, e2, 'r', label='e2')
    plt.plot(E, e1_l, 'b--', label='e1_l')
    plt.plot(E, e2_l, 'r--', label='e2_l')
    plt.legend()
    plt.show()



import numpy as np

import solcore.analytic_solar_cells as ASC
from solcore.light_source import LightSource
from solcore.state import State
from solcore.optics import solve_beer_lambert, solve_tmm, solve_rcwa, rcwa_options, solve_external_optics
from solcore.absorption_calculator import RCWASolverError
from solcore.structure import Layer, Junction, TunnelJunction

try:
    import solcore.poisson_drift_diffusion as PDD

    a = PDD.pdd_options
except AttributeError:
    PDD.pdd_options = {}

default_options = State()
pdd_options = PDD.pdd_options
asc_options = ASC.db_options

def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = State()
    for dictionary in dict_args:
        result.update(dictionary)
    return result


# General
default_options.T_ambient = 298
default_options.T = 298

# Illumination spectrum
default_options.wavelength = np.linspace(300, 1800, 251) * 1e-9
default_options.light_source = LightSource(source_type='standard', version='AM1.5g', x=default_options.wavelength,
                                           output_units='photon_flux_per_m')

# IV control
default_options.voltages = np.linspace(0, 1.2, 100)
default_options.mpp = False
default_options.light_iv = False
default_options.internal_voltages = np.linspace(-6, 4, 1000)
default_options.position = None
default_options.radiative_coupling = False

# Optics control
default_options.optics_method = 'BL'
default_options.recalculate_absorption = False

default_options = merge_dicts(default_options, ASC.db_options, PDD.pdd_options, rcwa_options)


def solar_cell_solver(solar_cell, task, user_options=None):
    """ Solves the properties of a solar cell object, either calculating its optical properties (R, A and T), its quantum efficiency or its current voltage characteristics in the dark or under illumination. The general options for the solvers are passed as dicionaries.

    :param solar_cell: A solar_cell object
    :param task: Task to perform. It has to be "optics", "iv", "qe", "equilibrium" or "short_circuit". The last two only work for PDD junctions
    :param user_options: A dictionary containing the options for the solver, which will overwrite the default options.
    :return: None
    """
    if type(user_options) in [State, dict]:
        options = merge_dicts(default_options, user_options)
    else:
        options = merge_dicts(default_options)

    prepare_solar_cell(solar_cell, options)
    options.T = solar_cell.T

    if task == 'optics':
        solve_optics(solar_cell, options)
    elif task == 'iv':
        solve_iv(solar_cell, options)
    elif task == 'qe':
        solve_qe(solar_cell, options)
    elif task == 'equilibrium':
        solve_equilibrium(solar_cell, options)
    elif task == 'short_circuit':
        solve_short_circuit(solar_cell, options)
    else:
        raise ValueError(
            'ERROR in "solar_cell_solver":\n\tValid values for "task" are: "optics", "iv", "qe", "equilibrium" and "short_circuit".')


def solve_optics(solar_cell, options):
    """ Solves the optical properties of the structure, calculating the reflectance, absorptance and transmitance. The "optics_method" option controls which method is used to calculate the optical properties of the solar cell:

    - None: The calculation is skipped. Only useful for solar cells involving just "2-diode" kind of junctions.
    - BL: Uses the Beer-Lambert law to calculate the absorption in each layer. Front surface reflexion has to provided externally. It is the default method and the most flexible one.
    - TMM: Uses a transfer matrix calculation to obtain the RAT. Not valid for DB or 2D junction
    - RCWA: Uses the rigorous wave coupled analysisto obtain the RAT. This allows to include 2D photonic crystals in the structure, for example. Not valid for DB or 2D junctions
    - external: The reflection and absorption profiles are provided externally by the user, and therefore no calculation is performed by Solcore.

    :param solar_cell: A solar_cell object
    :param options: Options for the optics solver
    :return: None
    """
    print('Solving optics of the solar cell...')

    calculated = hasattr(solar_cell[0], 'absorbed')
    recalc = options.recalculate_absorption if 'recalculate_absorption' in options.keys() else False
    if not calculated or recalc:

        if options.optics_method is None:
            print('Warning: Not solving the optics of the solar cell.')
        elif options.optics_method == 'external':
            solve_external_optics(solar_cell, options)
        elif options.optics_method == 'BL':
            solve_beer_lambert(solar_cell, options)
        elif options.optics_method == 'TMM':
            solve_tmm(solar_cell, options)
        elif options.optics_method == 'RCWA':
            if solve_rcwa is not None:
                solve_rcwa(solar_cell, options)
            else:
                raise RCWASolverError("RCWA optical solver not available!!")
        else:
            raise ValueError(
                'ERROR in "solar_cell_solver":\n\tOptics solver method must be None, "external", "BL", "TMM" or "RCWA".')

    else:
        print('Already calculated reflection, transmission and absorption profile - not recalculating. '
              'Set recalculate_absorption to True in the options if you want absorption to be calculated again.')


def solve_iv(solar_cell, options):
    """ Calculates the IV at a given voltage range, providing the IVs of the individual junctions in addition to the total IV

    :param solar_cell: A solar_cell object
    :param options: Options for the solvers
    :return: None
    """
    solve_optics(solar_cell, options)

    print('Solving IV of the junctions...')
    for j in solar_cell.junction_indices:

        if solar_cell[j].kind == 'PDD':
            PDD.iv_pdd(solar_cell[j], options)
        elif solar_cell[j].kind == 'DA':
            ASC.iv_depletion(solar_cell[j], options)
        elif solar_cell[j].kind == '2D':
            ASC.iv_2diode(solar_cell[j], options)
        elif solar_cell[j].kind == 'DB':
            ASC.iv_detailed_balance(solar_cell[j], options)
        else:
            raise ValueError(
                'ERROR in "solar_cell_solver":\n\tJunction {} has an invalid "type". It must be "PDD", "DA", "2D" or "DB".'.format(
                    j))

    print('Solving IV of the tunnel junctions...')
    for j in solar_cell.tunnel_indices:

        if solar_cell[j].kind == 'resistive':
            # The tunnel junction is modeled as a simple resistor
            ASC.resistive_tunnel_junction(solar_cell[j], options)
        elif solar_cell[j].kind == 'parametric':
            # The tunnel junction is modeled using a simple parametric model
            ASC.parametric_tunnel_junction(solar_cell[j], options)
        elif solar_cell[j].kind == 'external':
            # The tunnel junction is modeled using a simple parametric model
            ASC.external_tunnel_junction(solar_cell[j], options)
        elif solar_cell[j].kind == 'analytic':
            print('Sorry, the analytical tunnel junction model is not implemented, yet.')
        else:
            raise ValueError(
                'ERROR in "solar_cell_solver":\n\tTunnel junction {} has an invalid "type". It must be "parametric", "analytic", "external" or "resistive".'.format(
                    j))

    print('Solving IV of the total solar cell...')
    ASC.iv_multijunction(solar_cell, options)


def solve_qe(solar_cell, options):
    """ Calculates the QE of all the junctions

    :param solar_cell: A solar_cell object
    :param options: Options for the solvers
    :return: None
    """

    solve_optics(solar_cell, options)

    print('Solving QE of the solar cell...')
    for j in solar_cell.junction_indices:
        if solar_cell[j].kind == 'PDD':
            PDD.qe_pdd(solar_cell[j], options)
        elif solar_cell[j].kind == 'DA':
            ASC.qe_depletion(solar_cell[j], options)
        elif solar_cell[j].kind == '2D':
            # We solve this case as if it were DB. Therefore, to work it needs the same inputs in the Junction object
            wl = options.wavelength
            ASC.qe_detailed_balance(solar_cell[j], wl)
        elif solar_cell[j].kind == 'DB':
            wl = options.wavelength
            ASC.qe_detailed_balance(solar_cell[j], wl)
        else:
            raise ValueError(
                'ERROR in "solar_cell_solver":\n\tJunction {} has an invalid "type". It must be "PDD", "DA", "2D" or "DB".'.format(
                    j))


def solve_equilibrium(solar_cell, options):
    """ Uses the PDD solver to calculate the properties of all the all the junctions under equilibrium

    :param solar_cell: A solar_cell object
    :param options: Options for the solvers
    :return: None
    """
    for j in solar_cell.junction_indices:

        if solar_cell[j].kind == 'PDD':
            PDD.equilibrium_pdd(solar_cell[j], options)
        else:
            print('WARNING: Only PDD junctions can be solved in "equilibrium".')


def solve_short_circuit(solar_cell, options):
    """ Uses the PDD solver to calculate the properties of all the all the junctions under short circuit

    :param solar_cell: A solar_cell object
    :param options: Options for the solvers
    :return: None
    """

    solve_optics(solar_cell, options)

    for j in solar_cell.junction_indices:

        if solar_cell[j].kind == 'PDD':
            PDD.short_circuit_pdd(solar_cell[j], options)
        else:
            print('WARNING: Only PDD junctions can be solved in "short_circuit".')


def prepare_solar_cell(solar_cell, options):
    """ This function scans all the layers and junctions of the cell, calculating the relative position of each of them with respect the front surface (offset).
    This information will later be use by the optical calculators, for example. It also processes the 'position' option, which determines the spacing used if the
    solver is going to calculate depth-dependent absorption.

    :param solar_cell: A solar_cell object
    :param options: an options (State) object with user/default options
    :return: None
    """
    offset = 0
    layer_widths = []
    for j, layer_object in enumerate(solar_cell):

        # Independent layers, for example in a AR coating
        if type(layer_object) is Layer:
            layer_widths.append(layer_object.width)

        # Each Tunnel junctions can also have some layers with a given thickness.
        elif type(layer_object) is TunnelJunction:
            junction_width = 0
            for i, layer in enumerate(layer_object):
                junction_width += layer.width
                layer_widths.append(layer.width)
            solar_cell[j].width = junction_width

        # For each junction, and layer within the junction, we get the layer width.
        elif type(layer_object) is Junction:

            try:
                kind = solar_cell[j].kind
            except AttributeError as err:
                print('ERROR preparing the solar cell: Junction {} has no kind!'.format(j))
                raise err

            # This junctions will not, typically, have a width
            if kind in ['2D', 'DB']:
                layer_widths.append(1e-6)
                # 2D and DB junctions do not often have a width (or need it) so we set an arbitrary width
                if not hasattr(layer_object, 'width'):
                    solar_cell[j].width = 1e-6  # 1 µm

            else:
                junction_width = 0
                for i, layer in enumerate(layer_object):
                    layer_widths.append(layer.width)
                    junction_width += layer.width
                solar_cell[j].width = junction_width

        solar_cell[j].offset = offset
        offset += solar_cell[j].width

    solar_cell.width = offset

    process_position(solar_cell, options, layer_widths)



def process_position(solar_cell, options, layer_widths):
    """
    To control the depth spacing, the user can pass:
        - a vector which specifies each position (in m) at which the depth should be calculated
        - a single number which specifies the spacing (in m) to generate the position vector, e.g. 1e-9 for 1 nm spacing
        - a list of numbers which specify the spacing (in m) to be used in each layer. This list can have EITHER the length
        of the number of individual layers + the number of junctions in the cell object, OR the length of the total number of individual layers including layers inside junctions.

    :param solar_cell: a SolarCell object
    :param options: aan options (State) object with user/default options
    :param layer_widths: list of widths of the individual layers in the stack, treating the layers within junctions as individual layers
    :return: None
    """

    if options.position is None:
        options.position = [max(1e-10, width/5000) for width in layer_widths]

        layer_offsets = np.insert(np.cumsum(layer_widths), 0, 0)
        options.position = np.hstack([np.arange(layer_offsets[j],
                                                layer_offsets[j] + layer_width,
                                                options.position[j]) for j, layer_width
                                      in enumerate(layer_widths)])

    elif isinstance(options.position, int) or isinstance(options.position, float):
        options.position = np.arange(0, solar_cell.width, options.position)

    elif isinstance(options.position, list) or isinstance(options.position, np.ndarray):
        if len(options.position) == 1:
            options.position = np.arange(0, solar_cell.width, options.position[0])

        if len(options.position) == len(solar_cell):
            options.position = np.hstack([np.arange(layer_object.offset, layer_object.offset + layer_object.width, options.position[j]) for j, layer_object in enumerate(solar_cell)])

        elif len(options.position) == len(layer_widths):
            layer_offsets = np.insert(np.cumsum(layer_widths), 0, 0)
            options.position = np.hstack([np.arange(layer_offsets[j], layer_offsets[j] + layer_width, options.position[j]) for j, layer_width in enumerate(layer_widths)])



from collections import OrderedDict


class State(OrderedDict):
    """ This class defines a convent way of expanding the attributes of an object, usually fixed during the definition of
    the class. In this case, the class is just a dictionary - a special type of it - and attributes are expanded by
    adding new keys to it.
    """
    def __getattr__(self, name):
        # print ("***", name)
        if name in ["_OrderedDict__root", "_OrderedDict__map"]:
            return OrderedDict.__getattribute__(self, name)

        if name in self:
            return self[name]
        else:
            raise KeyError("The state object does not have an entry for the key '%s'." % (name,))

    def __setattr__(self, name, value):
        if name in ["_OrderedDict__root", "_OrderedDict__map", "_OrderedDict__hardroot"]:
            return OrderedDict.__setattr__(self, name, value)

        self[name] = value

# 
if __name__ == "__main__":
    a = State()

    a.a = 0
    a.b = 1
    a.c = 2

    a["d"] = 3
    print(a)



from solcore.structure import Layer, Junction, TunnelJunction

import numpy as np
import types


def solve_external_optics(solar_cell, options):
    """ Calculates the reflection, transmission and absorption of a solar cell from external data, preparing the structure for further calculations. The external data that must be provided in the solar cell definition is:

    external_reflected: a function that provides the fraction of reflected light at the specified wavelengths (in m)
    external_absorbed: a function that provides an array with the differential absorption at a depth Z (in m) at the specified wavelengths.

    Note that both are functions of a single variable - wavelength in the first case and position in the second - and that the latter provides as output an array at the wavelengths indicated in the options.

    :param solar_cell:
    :param options:
    :return:
    """
    solar_cell.wavelength = options.wavelength

    # We include the shadowing losses
    initial = (1 - solar_cell.shading) if hasattr(solar_cell, 'shading') else 1

    # We try to get the external attributes
    try:
        solar_cell.reflected = solar_cell.external_reflected * initial
        diff_absorption = solar_cell.external_absorbed
    except AttributeError as err:
        raise err

    # We calculate the total amount of light absorbed in the solar cell, integrating over its whole thickness with a step of 1 nm
    all_z = options.position
    all_absorbed = np.trapz(diff_absorption(all_z), all_z)

    # Each building block (layer or junction) needs to have access to the absorbed light in its region.
    # We update each object with that information.
    for j in range(len(solar_cell)):
        solar_cell[j].diff_absorption = diff_absorption
        solar_cell[j].absorbed = types.MethodType(absorbed, solar_cell[j])

    solar_cell.transmitted = (1 - solar_cell.external_reflected - all_absorbed) * initial
    solar_cell.absorbed = all_absorbed * initial


def absorbed(self, z):
    out = self.diff_absorption(self.offset + z) * (z < self.width)
    return out.T



import re
import numpy
from configparser import ConfigParser
import os
import matplotlib

matplotlib.use("TkAgg")
import matplotlib.pyplot as plt




interactive_terminal = False

float_re = re.compile(r"[-+]?[0-9]*\.?[0-9]+")

class CPPicker:
    """ This class can be used to pick the critical points out of a collection of n and k curves for a range of
    compositions in ternary alloys. Later, this critical points can be used later to calculate the n and k data at any
    other intermediate composition.

    >>> directory = '/InGaSb-Material/k.txt'
    >>> pickerApp = CPPicker(directory)
    >>> pickerApp.loop()

    """
    def __init__(self, directory):
        self.current_critical_point = 0
        self.current_curve = 0
        self.directory = directory
        self.gather_files()
        self.fraction_order = list(sorted(self.fractions_dict.keys()))
        self.n_curves = len(self.fraction_order)
        self.critical_points = [[]]
        self.finished = False
        self.log = True
        self.snap_to_data = True

    def guess_fraction(self, file_name):
        re_matches = float_re.findall(file_name)
        return re_matches[0] if len(re_matches) != 0 else None

    def gather_files(self):
        full_directory_path = self.directory
        candidate_files = [
            os.path.join(directory, f)
            for f in os.listdir(full_directory_path)
            if ".txt" in f and "critical" not in f
            ]
        results = {}
        for file_path in sorted(candidate_files):
            _, file_name = os.path.split(file_path)
            fraction_guess = self.guess_fraction(file_name)
            print("File '{}': material fraction guess = {} {}".format(
                file_name,
                fraction_guess,
                ("OK?\n> " if interactive_terminal else "\n")
            ), end="")

            answer = input() if interactive_terminal else ""
            fraction = fraction_guess if answer == "" else answer

            if interactive_terminal:
                print("using: {}".format(fraction))

            file_info = {
                "file_path": file_path,
                "file_name": file_name,
                "fraction": fraction,
                "data": numpy.loadtxt(file_path, unpack=True)
            }
            results[fraction] = file_info
        self.fractions_dict = results

    def plot_a_curve(self, curve_id, background=False):

        data = self.fractions_dict[curve_id]["data"]

        if (curve_id == self.fraction_order[self.current_curve] and not background):
            colour = "red"
            picker = 1000
            width = 3
        elif (self.current_curve != 0 and curve_id == self.fraction_order[self.current_curve - 1]):
            colour = "black"
            picker = None
            width = 1
        else:
            colour = "lightgrey"
            picker = None
            width = 1

        plt.plot(data[0] * 1e9, data[1], "-", color=colour, picker=picker, linewidth=width)

    def setup_plot(self, keep_limits=False):
        self.fig = plt.figure()
        self.fig.canvas.mpl_connect('pick_event', self.onpick)
        self.fig.canvas.mpl_connect('key_press_event', self.onkey)
        plt.ion()
        self.plot_all(keep_limits=False)

    def plot_all(self, keep_limits=True):
        if keep_limits:
            ranges = plt.xlim(), plt.ylim()
            self.fig.clear()
            plt.xlim(*ranges[0])
            plt.ylim(*ranges[1])
        else:
            self.fig.clear()
        if self.log:
            plt.yscale("log")
        else:
            plt.yscale("linear")

        for fraction in self.fractions_dict:
            self.plot_a_curve(fraction, background=True)
        self.plot_a_curve(self.fraction_order[self.current_curve], background=False)

        x_old = []
        y_old = []

        for critical_point_sublist in self.critical_points:
            x, y = [point[0] for point in critical_point_sublist], [point[1] for point in critical_point_sublist]
            plt.plot(x, y, ":+", color="blue", markersize=10, )

        plt.draw()
        plt.show()

    def loop(self):
        current_critical_point = 0
        self.setup_plot()

        while self.finished == False:
            self.fig.canvas.get_tk_widget().update()

        self.save()

    def onpick(self, event):

        thisline = event.artist
        xdata = thisline.get_xdata()
        ydata = thisline.get_ydata()
        ind = event.ind
        x = event.mouseevent.xdata
        y = numpy.interp(x, xdata, ydata) if self.snap_to_data else event.mouseevent.ydata

        if self.current_curve < self.n_curves:
            if len(self.critical_points[-1]) < self.n_curves:
                print("critical for fraction {} at: x,y=({},{})".format(self.fraction_order[self.current_curve], x, y))
                self.critical_points[-1].append((x, y))

        if self.current_curve < self.n_curves - 1:
            self.current_curve += 1

        self.plot_all()

    def onkey(self, event):
        print("event.key", event.key)
        if event.key == "backspace":
            if len(self.critical_points[-1]) != 0:
                self.critical_points[-1].pop(-1)
                self.current_curve = len(self.critical_points[-1])
        if event.key == "enter":
            if len(self.critical_points[-1]) != self.n_curves:
                print("Insufficient critical points, continuing...")
            else:
                self.critical_points.append(list())
                self.current_curve = 0
        if event.key == "escape":
            self.finished = True
        if event.key == 'l':
            self.log = not self.log
        if event.key == 'n':
            self.snap_to_data = not self.snap_to_data

        self.plot_all()

    def save(self):

        out_path = os.path.join(self.directory, "temp_critical_points.txt")
        print("output path: {}".format(out_path))
        if interactive_terminal:
            print("chante_to? >")

        out_path = input() if interactive_terminal else out_path
        if interactive_terminal:
            print("using: {}".format(out_path))

        self.result_config = ConfigParser()

        for section in self.fraction_order:
            self.result_config.add_section(section)
            self.result_config.set(section, "file", self.fractions_dict[section]["file_name"])

        for critical_point_number, critical_points_subset in enumerate(self.critical_points[:-1]):
            for fraction_number, fraction in enumerate(self.fraction_order):
                c = critical_points_subset[fraction_number]
                self.result_config.set(fraction, str(critical_point_number + 1), "{} {}".format(c[0] * 1e-9, c[1]))

        with open(out_path, "w") as f:
            self.result_config.write(f)


if __name__ == '__main__':
    directory = '/Users/diego/OneDrive - Imperial College London/Developement/solcore5/solcore/material_data/AlInP-Material/k'
    pickerApp = CPPicker(directory)
    pickerApp.loop()



# Custom_CPPB module for building customisable optical constant models based on any number of local oscillator functions

import numpy as np
import os, sys
import re
from solcore.science_tracker import science_reference
import solcore.constants as const


class Custom_CPPB():
    """
    Customisable Critical-Point Parabolic-Band model (CPPB) using expressions from Sadao Adachi and Charles Kim.

    The Custom_CPPB() class is designed to work with solcore's structure module, where now you can define a structure
    containing any number of oscillator functions and calculate the corresponding optical constants.

    """

    def __init__(self):

        # Define filepath to the MaterialParameters file import...
        DIR, SCRIPT = os.path.split(__file__)

        self.__PARAMS_PATH = os.path.join(DIR, 'Custom_CPPB_MaterialParamaters.txt')

        # Opens MaterialParameters file and dumps all content into a list...
        with open(self.__PARAMS_PATH, mode="r") as File:
            self.__MatParams = File.read().splitlines()

        # Single line lists all callable functions in the Custom_CPPB class, with the exception of __init__...
        self.__METHODS = [func for func in dir(Custom_CPPB)
                          if callable(getattr(Custom_CPPB, func)) and not func.startswith("__")]

    def Material_Params(self, material, *parameters):
        """
        Custom_CPPB.Material_Params() :: loads the required material parameters from the database file and
            returns a dictionary of all material parameters or individual variable for the required material.

        :param material: string variable, must match the available materials in Mod_CPPB_MaterialParameters.txt
        :param parameters: variable length argument to specify individual parameters if required.
                            Must be entered as strings.

        :return: Material Parameters
        """
        # Initialise variables
        ID = None
        Found = False
        PARAMS = {}
        PARAMS.setdefault("Material", material)

        for Lines in self.__MatParams:
            # Using regular expressions package, matching the material name with the correct section of the file.
            # If the name is found, ID is switched to True.
            if re.fullmatch("Material :: " + material, Lines) != None:
                ID = True
                Found = True

            # if the $END$ identifier is found in a line the ID is switched to False
            if "$END$" in Lines:
                ID = False

            # Whilst ID is true the parameters are collected and stored in a dictionary variable.
            # The additional conditional statements tell the code to only store 'variable' lines (containing =) and to ignore
            # comment lines (containing #).
            if ID is True and "=" in Lines and "$" not in Lines:
                split_line = Lines.split(" = ")

                PARAMS[split_line[0]] = float(split_line[1])

        if Found is False:
            raise ValueError("Material not found...")

        # If specific parameter or parameters are required, check to see if it exists and if so return it.
        if parameters.__len__() == 0:

            return PARAMS

        else:

            PARAMS_LIST = {}

            for p in parameters:

                if p in PARAMS:
                    PARAMS_LIST[p] = PARAMS[p]

                else:
                    raise ValueError("Material_Params ERROR :: Material parameter %s not found..." % p)

            return PARAMS_LIST

    def Broad(self, Gamma, Alpha, E0, energy):
        """
        Custom_CPPB.Broad() :: defines the frequency dependent gaussian broadening function proposed by C. Kim.

        :param Gamma: Broadening parameter (eV).
        :param Alpha: Parameter describing the transition from pure Lorentzian (Alpha=0) to approximated Gaussian
                (Alpha = 0.2) lineshape broadneing.
        :param E0: Critical point centre energy (eV).
        :param energy: Energy array (eV).

        :return: Frequency dependent broadening parameter.
        """

        # Scientific reference for this work...
        science_reference("Charles Kim, 'Modelling the optical dielectric function of semiconductors",
                          "C. C. Kim et al, 'Modelling the optical dielectric function of semiconductors: Extension of "
                          "the critical-point parabolic band approximation', Physical Review B 45(20) 11749, 1992")

        return Gamma * np.exp(-1 * Alpha * ((energy - E0) / Gamma) ** 2)

    def E0andE0_d0(self, energy, material_parameters=None, **kwargs):
        """
        Custom_CPPB.E0andE0_d0() = Intraband transition region, E0 and E0+delta0 transitions at the 3D M0 Critical
            Point.

        :param energy: energy: Energy array (eV).
        :param material_parameters: Parameter set imported using Material_Parameters() method. Not required as long as
            keyword arguments are specified.
        :param kwargs: These take in individual parameters for the model. Keywords should take the following form;
                    Gamma_E0
                    Alpha_E0
                    E0
                    E0_d0
                    A

        :return: E0 and E0_d0 critical point contributions to the complex dielectric function.
        """

        # Scientific reference for this work...
        science_reference("Sadao Adachi, Physical Properties of III-V Semiconductor Compounds",
                          "Adachi, S., Physical Properties of III-V Semiconductor Compounds, John Wiley & Sons (1992)")

        # Conditional statement determining where the input parameters come from...
        if material_parameters is None and bool(kwargs) is False:

            raise ValueError("No material parameters specified...")

        elif material_parameters is not None and bool(kwargs) is False:

            Params = material_parameters

        elif material_parameters is not None and bool(kwargs) is True:

            Params = material_parameters

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        elif bool(kwargs) is True:

            Params = {}

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        # Frequency dependent broadening...
        Gamma = self.Broad(Params["Gamma_E0"], Params["Alpha_E0"], Params["E0"], energy)

        # Oscillator model based on the form of an 3D M0 type critical point...
        ChiO = (energy + 1j * Gamma) / Params["E0"]
        ChiSO = (energy + 1j * Gamma) / Params["E0_d0"]

        F_ChiO = (ChiO ** (-2)) * (2 - (1 + ChiO) ** 0.5 - (1 - ChiO) ** 0.5)
        F_ChiSO = (ChiSO ** (-2)) * (2 - (1 + ChiSO) ** 0.5 - (1 - ChiSO) ** 0.5)

        Eps = (Params["A"] * (Params["E0"] ** (-1.5))) * (F_ChiO + 0.5 * ((Params["E0"] / Params["E0_d0"]) ** 1.5) \
                                                          * F_ChiSO)

        # Additional line to address change in phase of the imaginary signal.
        return Eps.real + 1j * abs(Eps.imag)

    def E1andE1_d1(self, energy, material_parameters, **kwargs):
        """
        Custom_CPPB.E1andE1_d1() = function describing the E1 and E1 + delta1 intraband transitions at the 3D M1 critical
            point.

        :param energy: energy: Energy array (eV).
        :param material_parameters: Parameter set imported using Material_Parameters() method. Not required as long as
            keyword arguments are specified.
        :param kwargs: These take in individual parameters for the model. Keywords should take the following form;
                    Gamma_E1
                    Alpha_E1
                    E1
                    E1_d1
                    B1
                    B1s

        :return: E1 and E1_d1 critical point contributions to the complex dielectric function.
        """

        # Scientific reference for this work...
        science_reference("Sadao Adachi, Physical Properties of III-V Semiconductor Compounds",
                          "Adachi, S., Physical Properties of III-V Semiconductor Compounds, John Wiley & Sons (1992)")

        # Conditional statement determining where the input parameters come from...
        if material_parameters is None and bool(kwargs) is False:

            raise ValueError("No material parameters specified...")

        elif material_parameters is not None and bool(kwargs) is False:

            Params = material_parameters

        elif material_parameters is not None and bool(kwargs) is True:

            Params = material_parameters

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        elif bool(kwargs) is True:

            Params = {}

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        # Frequency dependent broadening parameter...
        Gamma = self.Broad(Params["Gamma_E1"], Params["Alpha_E1"], Params["E1"], energy)

        # Oscillator function based on the form of an 2D M0 type critical point...
        Chi_1D = (energy + 1j * Gamma) / Params["E1"]
        Chi_1SD = (energy + 1j * Gamma) / Params["E1_d1"]

        Eps = (-1 * Params["B1"] * Chi_1D ** (-2)) * np.log(1 - Chi_1D ** 2) + (-1 * Params["B1s"] * Chi_1SD ** (-2)) \
                                                                               * np.log(1 - Chi_1SD ** 2)

        # Additional line to address change in phase of the imaginary signal.
        return Eps.real + 1j * abs(Eps.imag)

    def E_ID(self, energy, material_parameters, **kwargs):
        """
        Custom_CPPB.Eg_ID() :: From Adachi's formalism, contributions to the complex dielectric function from the indirect
            band-gap transitions.

        :param energy: energy: Energy array (eV).
        :param material_parameters: Parameter set imported using Material_Parameters() method. Not required as long as
            keyword arguments are specified.
        :param kwargs: These take in individual parameters for the model. Keywords should take the following form;
                    Gamma_Eg_ID
                    Alpha_Eg_ID
                    Eg_ED
                    Ec
                    D

        :return: Indirect band gap contributions to the complex dielectric function.
        """

        # Scientific reference for this work...
        science_reference("Sadao Adachi, Physical Properties of III-V Semiconductor Compounds",
                          "Adachi, S., Physical Properties of III-V Semiconductor Compounds, John Wiley & Sons (1992)")

        # Conditional statement determining where the input parameters come from...
        if material_parameters is None and bool(kwargs) is False:

            raise ValueError("No material parameters specified...")

        elif material_parameters is not None and bool(kwargs) is False:

            Params = material_parameters

        elif material_parameters is not None and bool(kwargs) is True:

            Params = material_parameters

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        elif bool(kwargs) is True:

            Params = {}

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        # Frequency dependent broadening parameter...
        Gamma = self.Broad(Params["Gamma_Eg_ID"], Params["Alpha_Eg_ID"], Params["Eg_ID"], energy)

        # Function describing the contributions of indirect transitions...
        term1 = -1 * (((Params["Eg_ID"] ** 2) / (energy + 1j * Gamma) ** 2) * np.log(Params["Ec"] / Params["Eg_ID"]))
        term2 = 0.5 * ((1 + (Params["Eg_ID"] / (energy + 1j * Gamma))) ** 2) * \
                np.log((energy + 1j * Gamma + Params["Ec"]) / (energy + 1j * Gamma + Params["Eg_ID"]))
        term3 = 0.5 * ((1 - (Params["Eg_ID"] / (energy + 1j * Gamma))) ** 2) * \
                np.log((energy + 1j * Gamma - Params["Ec"]) / (energy + 1j * Gamma - Params["Eg_ID"]))

        Eps = (2 * Params["D"] / np.pi) * (term1 + term2 + term3)

        # Additional line to address change in phase of the imaginary signal.
        return Eps.real + 1j * abs(Eps.imag)

    def E0_Exciton(self, energy, material_parameters, **kwargs):
        """
        Custom_CPPB.E0_Exciton() :: From Adachi's formalism, contributions to the complex dielectric function from bound
            excitons in the vicinity of the E0 fundamental band-gap.

        :param energy: energy: Energy array (eV).
        :param material_parameters: Parameter set imported using Material_Parameters() method. Not required as long as
            keyword arguments are specified.
        :param kwargs: These take in individual parameters for the model. Keywords should take the following form;
                    Gamma_Ex0
                    Alpha_Ex0
                    A_Ex
                    G_3D
                    n

        :return: Excitonic contributions at E0 to the complex dielectric function.
        """

        # Scientific reference for this work...
        science_reference("Sadao Adachi, Physical Properties of III-V Semiconductor Compounds",
                          "Adachi, S., Physical Properties of III-V Semiconductor Compounds, John Wiley & Sons (1992)")

        # Conditional statement determining where the input parameters come from...
        if material_parameters is None and bool(kwargs) is False:

            raise ValueError("No material parameters specified...")

        elif material_parameters is not None and bool(kwargs) is False:

            Params = material_parameters

        elif material_parameters is not None and bool(kwargs) is True:

            Params = material_parameters

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        elif bool(kwargs) is True:

            Params = {}

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        # Frequency dependent broadening parameter...
        Gamma = self.Broad(Params["Gamma_Ex0"], Params["Alpha_Ex0"], Params["E0"], energy)
        Eps_n = np.zeros((len(energy), int(Params["n"] + 1)), dtype="complex")

        for j in range(1, int(Params["n"] + 1)):
            Eps_n[:, j] = (Params["A_Ex"] / j ** 3) * ((Params["E0"] - (Params["G_3D"] / \
                                                                        (j ** 2)) - energy - 1j * Gamma) ** -1)

        Eps = np.sum(Eps_n, 1)

        # Additional line to address change in phase of the imaginary signal.
        return Eps.real + 1j * abs(Eps.imag)

    def E2(self, energy, material_parameters, **kwargs):
        """
        Custom_CPPB.E2() :: S. Adachi finds that the high energy E2 critical point is well approximated using a simple
            damped harmonic oscillator.

        :param energy: energy: Energy array (eV).
        :param material_parameters: Parameter set imported using Material_Parameters() method. Not required as long as
            keyword arguments are specified.
        :param kwargs: These take in individual parameters for the model. Keywords should take the following form;
                    Gamma_E2
                    Alpha_E2
                    E2
                    C

        :return: E2 contributions to the complex dielectric function.
        """

        # Scientific reference for this work...
        science_reference("Sadao Adachi, Physical Properties of III-V Semiconductor Compounds",
                          "Adachi, S., Physical Properties of III-V Semiconductor Compounds, John Wiley & Sons (1992)")

        # Conditional statement determining where the input parameters come from...
        if material_parameters is None and bool(kwargs) is False:

            raise ValueError("No material parameters specified...")

        elif material_parameters is not None and bool(kwargs) is False:

            Params = material_parameters

        elif material_parameters is not None and bool(kwargs) is True:

            Params = material_parameters

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        elif bool(kwargs) is True:

            Params = {}

            for key in kwargs:

                try:

                    Params[key] = kwargs[key]
                except KeyError:
                    print("Invalid material parameter...")

        # Frequency dependent broadning parameter...
        Gamma = self.Broad(Params["Gamma_E2"], Params["Alpha_E2"], Params["E2"], energy)

        # Damped harmonic oscillator function described by Adachi...
        Eps = Params["C"] / ((1 - (energy / Params["E2"]) ** 2) - 1j * (energy / Params["E2"]) * Gamma)

        # Additional line to address change in phase of the imaginary signal.
        return Eps.real + 1j * abs(Eps.imag)

    def Lorentz(self, energy, **kwargs):
        """
        Custom_CPPB.Lorentz() :: Classic Lorentz oscillator expression, taken from the J. A. Woollam ellipsometer
            manual.

        :param energy: Energy array (eV)
        :param kwargs: These take in individual parameters for the model. Keywords should take the following form;
                    Gamma
                    Alpha
                    E0
                    Amp

        :return: Lorentz oscillator contributions to the complex dielectric function.
        """

        # Scientific reference for this work...
        science_reference("J. A. Woollam, Guide to using WVASE", "J. A. Woollam Co. Inc., Copyright 1994-2012")

        # Remove material_parameters argument from the kwargs dictionary to avoid confusion...
        if "material_parameters" in kwargs:
            del kwargs["material_parameters"]

        Gamma = self.Broad(kwargs["Gamma"], kwargs["Alpha"], kwargs["E0"], energy)

        # Eps = (Params["C"] * Params["E0"]**2) / ((Params["E0"]**2 - energy**2) - 1j*energy*Gamma)
        Eps = (kwargs["Amp"] * kwargs["E0"]) / (kwargs["E0"] ** 2 - energy ** 2 - 1j * Gamma * energy)

        # Additional line to address change in phase of the imaginary signal.
        return Eps.real + 1j * abs(Eps.imag)

    def Sellmeier(self, energy, **kwargs):
        """
        Custom_CPPB.Sellmeier() :: Calculate the Sellmeier dispersion relation for the entirely real parts of the
            dielectric function. The Sellmeier relation is a sum of N terms.

        :param energy: Energy array (eV)
        :param kwargs: These take in individual parameters for the model. Parameters should be in units of "um".
            Keywords should take the following form;
                    An
                    Ln
                    NOTE: 'n' should take an integer from 1 to N where N is the number of Sellmeier terms.

        :return: Sellmeier contributions to the real part of the dielectric function.
        """

        # Scientific reference for this work...
        science_reference("Wilhelm Sellmeier's development of the Cauchy dispersion relation, taken from the J."
                          " A. Woollam WVASE manual", "J. A. Woollam Co. Inc., Copyright 1994-2012")

        # Remove material_parameters argument from the kwargs dictionary to avoid confusion...
        if "material_parameters" in kwargs:
            del kwargs["material_parameters"]

        # Work out number of Sellmeier terms from length of kwargs...
        if np.mod(len(kwargs), 2) != 0:

            raise ValueError("Insufficient number of Sellmeier coefficients :: No. kwargs == %d" % len(kwargs))

        else:

            terms = len(kwargs) / 2

        # Define empty array for individual Sellmeier terms...
        epsilon = np.zeros((int(terms), len(energy)), dtype=complex)

        # define the conversion constant to go between energy in eV and lambda in SI units...
        C = ((const.h * const.c) / const.q) * 1E6

        # calculate Sellmeier expression for the given coefficients...
        for i in range(0, int(terms)):
            epsilon[i] = (kwargs["A%d" % (i + 1)] * (C / energy) ** 2) / (
            (C / energy) ** 2 - kwargs["L%d" % (i + 1)] ** 2)

        return sum(epsilon, 0) + 1

    # def eps_calc(self, oscillator_structure, energy_array, components=False):
    #     """
    #     Custom_CPPB.eps_calc() :: Calculates the complex dielectric function of the presented oscillator structure.
    #
    #     :param oscillator_structure: Structure object containing information about the individual component functions.
    #     :param energy_array: Energy array (eV)
    #     :param components: Default=False, selects whether components are output along with the final result.
    #
    #     :return: Complex dielectric function and components from each oscillator function.
    #     """
    #
    #     # Statement checking whether the components argument has been specified as boolean...
    #     if isinstance(components, bool) == False:
    #         raise ValueError("'components' variable is invalid... state as either 'True or False")
    #
    #     # Find the length of oscillator structure and energy array and initialise a complex array for epsilon...
    #     epsilon = np.zeros((len(oscillator_structure), len(energy_array)), dtype="complex")
    #
    #     # Calculate the contributions from each specified oscillator...
    #     ind = 0
    #     for Osc_Num in oscillator_structure:
    #
    #         if Osc_Num.oscillator == "E0andE0_d0":
    #
    #             epsilon[ind] = self.E0andE0_d0(energy=energy_array, material_parameters=Osc_Num.material_parameters,
    #                                              **Osc_Num.arguments)
    #         elif Osc_Num.oscillator == "E1andE1_d1":
    #
    #             epsilon[ind] = self.E1andE1_d1(energy=energy_array, material_parameters=Osc_Num.material_parameters,
    #                                               **Osc_Num.arguments)
    #         elif Osc_Num.oscillator == "E_ID":
    #
    #             epsilon[ind] = self.E_ID(energy=energy_array, material_parameters=Osc_Num.material_parameters,
    #                                               **Osc_Num.arguments)
    #         elif Osc_Num.oscillator == "E0_Exciton":
    #
    #             epsilon[ind] = self.E_ID(energy=energy_array, material_parameters=Osc_Num.material_parameters,
    #                                         **Osc_Num.arguments)
    #         elif Osc_Num.oscillator == "E2":
    #
    #             epsilon[ind] = self.E2(energy=energy_array, material_parameters=Osc_Num.material_parameters,
    #                                      **Osc_Num.arguments)
    #         elif Osc_Num.oscillator == "Lorentz":
    #
    #             epsilon[ind] = self.Lorentz(energy=energy_array, **Osc_Num.arguments)
    #         else:
    #
    #             raise ValueError("Custom_CPPB() does not contain an oscillator function with name = " +
    #                              Osc_Num.oscillator)
    #
    #         ind += 1
    #
    #     # Sum all the individual contributions to produce the total complex dalectric function. The eps_inf parameter
    #     # is also added to the data.
    #     try:
    #         eps1_inf = oscillator_structure[0].material_parameters["eps1_inf"]
    #     except:
    #         eps1_inf = 0.0
    #
    #     epsilon_tot = sum(epsilon, 0) + eps1_inf
    #
    #     # Select whether component outputs are required...
    #     if components == False:
    #
    #         return epsilon_tot
    #
    #     elif components == True:
    #
    #         return (epsilon_tot, epsilon)

    def eps_calc(self, oscillator_structure, energy_array):
        """
        Custom_CPPB.eps_calc() :: Calculates the complex dielectric function of the presented oscillator structure.

        :param oscillator_structure: Structure object containing information about the individual component functions.
        :param energy_array: Energy array (eV)

        :return: A dictionary containing the complex dielectric function and components from each oscillator function.
            "eps" :: Sum of all contributions.
            "components" :: Individual components in a numpy array
        """

        # Find the length of oscillator structure and energy array and initialise a complex array for epsilon...
        epsilon = np.zeros((len(oscillator_structure), len(energy_array)), dtype="complex")

        # Calculate the contributions from each specified oscillator...
        i = 0

        for Osc in oscillator_structure:

            try:

                getattr(self, Osc.oscillator)
            except KeyError:

                print("Custom_CPPB() does not contain an oscillator function with name = " + Osc.oscillator)

            method = getattr(self, Osc.oscillator)

            epsilon[i] = method(energy=energy_array, material_parameters=Osc.material_parameters,
                                **Osc.arguments)

            i += 1

        # Sum all the individual contributions to produce the total complex dalectric function. The eps_inf parameter
        # is also added to the data.
        try:

            eps1_inf = oscillator_structure[0].material_parameters["eps1_inf"]
        except:

            eps1_inf = 0.0

        epsilon_tot = sum(epsilon, 0) + eps1_inf

        # A dictionary containing the total of all contributions to epsilon and also the components...
        return {"eps": epsilon_tot, "components": epsilon}

    def alpha_calc(self, oscillator_structure, energy_array):
        """
        Custom_CPPB.alpha_calc() :: Calculates the absorption coefficient from the complex dielectric function.

        :param oscillator_structure: Structure object containing information about the individual component functions.
        :param energy_array: Energy array (eV)

        :return: A dictionary containing absorption coeffcient and components from individual oscillator functions.
            "alpha" :: returns the total contributions to the absorption coefficient.
            "components" :: returns the individual components of alpha from each oscillator function.
        """

        # Define lambda functions for calculating k and alpha quickly...
        k_func = lambda Epsilon: np.sqrt(((Epsilon.real ** 2 + Epsilon.imag ** 2) ** 0.5 - Epsilon.real) / 2)
        alpha_func = lambda energy, k: ((4 * const.pi) / ((const.h * const.c) / (energy * const.q))) * k

        # Calculate the complex dielectruc function...
        epsilon = self.eps_calc(oscillator_structure, energy_array)

        # Calculate alpha for each component...
        Alpha_components = alpha_func(energy_array, k_func(epsilon["components"]))
        Alpha = alpha_func(energy_array, k_func(epsilon["eps"]))

        return {"alpha": Alpha, "components": Alpha_components}

    def nk_calc(self, oscillator_structure, energy_array):
        """
        Custom_CPPB.nk_calc() :: Calculates the refractive index and extinction coefficient from the complex dielectric
            function.

        :param oscillator_structure: Structure object containing information about the individual component functions.
        :param energy_array: Energy array (eV)

        :return: A dictionary containing refractive index, extinction coefficient and their components from individual
            oscillator functions.
            "n" :: returns refractive index.
            "n_components" :: returns components of refractive index from individual oscillator functions.
            "k" :: returns extinction coefficient.
            "k_components" :: returns components of extinction coefficient from individual oscillator functions.
        """

        # Define lambda function for calculating n...
        n_func = lambda Epsilon: np.sqrt(((Epsilon.real ** 2 + Epsilon.imag ** 2) ** 0.5 + Epsilon.real) / 2)
        # Define lambda function for calculating k...
        k_func = lambda Epsilon: np.sqrt(((Epsilon.real ** 2 + Epsilon.imag ** 2) ** 0.5 - Epsilon.real) / 2)

        # Calculate the complex dielectruc function...
        epsilon = self.eps_calc(oscillator_structure, energy_array)

        # Calculate n for each component...
        n_components = n_func(epsilon["components"])
        n = n_func(epsilon["eps"])

        # Calculate k for each component...
        k_components = k_func(epsilon["components"])
        k = k_func(epsilon["eps"])

        return {"n": n, "n_components": n_components, "k": k, "k_components": k_components}
